from typing import Type

from langchain_core.documents import Document
from langchain_tests.integration_tests import (
    RetrieversIntegrationTests,
)
from pymongo import MongoClient
from pymongo.collection import Collection

from langchain_mongodb import MongoDBAtlasVectorSearch
from langchain_mongodb.index import (
    create_fulltext_search_index,
)
from langchain_mongodb.retrievers import (
    MongoDBAtlasFullTextSearchRetriever,
    MongoDBAtlasHybridSearchRetriever,
)

from ..utils import (
    CONNECTION_STRING,
    DB_NAME,
    TIMEOUT,
    ConsistentFakeEmbeddings,
    PatchedMongoDBAtlasVectorSearch,
)

DIMENSIONS = 5
COLLECTION_NAME = "langchain_test_retrievers_standard"
VECTOR_INDEX_NAME = "vector_index"
PAGE_CONTENT_FIELD = "text"
SEARCH_INDEX_NAME = "text_index"


def get_collection() -> Collection:
    client = MongoClient(CONNECTION_STRING)
    return client[DB_NAME][COLLECTION_NAME]


def setup_test(coll: Collection) -> MongoDBAtlasVectorSearch:
    # Set up the vector search index and add the documents if needed.
    vs = PatchedMongoDBAtlasVectorSearch(
        coll,
        embedding=ConsistentFakeEmbeddings(DIMENSIONS),
        dimensions=DIMENSIONS,
        index_name=VECTOR_INDEX_NAME,
        text_key=PAGE_CONTENT_FIELD,
        auto_index_timeout=TIMEOUT,
    )

    if coll.count_documents({}) == 0:
        vs.add_documents(
            [
                Document(page_content="In 2023, I visited Paris"),
                Document(page_content="In 2022, I visited New York"),
                Document(page_content="In 2021, I visited New Orleans"),
                Document(page_content="Sandwiches are beautiful. Sandwiches are fine."),
            ]
        )

    # Set up the search index if needed.
    if not any([ix["name"] == SEARCH_INDEX_NAME for ix in coll.list_search_indexes()]):
        create_fulltext_search_index(
            collection=coll,
            index_name=SEARCH_INDEX_NAME,
            field=PAGE_CONTENT_FIELD,
            wait_until_complete=TIMEOUT,
        )

    return vs


class TestMongoDBAtlasFullTextSearchRetriever(RetrieversIntegrationTests):
    _coll: Collection

    @classmethod
    def setup_class(cls):
        cls._coll = get_collection()

    @classmethod
    def teardown_class(cls):
        cls._coll.database.client.close()

    @property
    def retriever_constructor(self) -> Type[MongoDBAtlasFullTextSearchRetriever]:
        """Get a retriever for integration tests."""
        return MongoDBAtlasFullTextSearchRetriever

    @property
    def retriever_constructor_params(self) -> dict:
        setup_test(self._coll)
        return {
            "collection": self._coll,
            "search_index_name": SEARCH_INDEX_NAME,
            "search_field": PAGE_CONTENT_FIELD,
        }

    @property
    def retriever_query_example(self) -> str:
        """
        Returns a str representing the "query" of an example retriever call.
        """
        return "When was the last time I visited new orleans?"


class TestMongoDBAtlasHybridSearchRetriever(RetrieversIntegrationTests):
    _coll: Collection

    @classmethod
    def setup_class(cls):
        cls._coll = get_collection()

    @classmethod
    def teardown_class(cls):
        cls._coll.database.client.close()

    @property
    def retriever_constructor(self) -> Type[MongoDBAtlasHybridSearchRetriever]:
        """Get a retriever for integration tests."""
        return MongoDBAtlasHybridSearchRetriever

    @property
    def retriever_constructor_params(self) -> dict:
        vs = setup_test(self._coll)
        return {
            "vectorstore": vs,
            "collection": self._coll,
            "search_index_name": SEARCH_INDEX_NAME,
            "search_field": PAGE_CONTENT_FIELD,
        }

    @property
    def retriever_query_example(self) -> str:
        """
        Returns a str representing the "query" of an example retriever call.
        """
        return "When was the last time I visited new orleans?"



"Demonstrates MongoDBAtlasVectorSearch.as_retriever() invoked in a chain"

from __future__ import annotations

import os

import pytest  # type: ignore[import-not-found]
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.output_parsers.string import StrOutputParser
from langchain_core.prompts.chat import ChatPromptTemplate
from langchain_core.runnables import RunnablePassthrough
from langchain_openai import AzureChatOpenAI, ChatOpenAI
from langchain_openai.chat_models.base import BaseChatOpenAI
from pymongo import MongoClient
from pymongo.collection import Collection

from langchain_mongodb import index

from ..utils import DB_NAME, PatchedMongoDBAtlasVectorSearch

COLLECTION_NAME = "langchain_test_chain_example"
INDEX_NAME = "langchain-test-chain-example-vector-index"
DIMENSIONS = 1536
TIMEOUT = 60.0
INTERVAL = 0.5


@pytest.fixture(scope="module")
def collection(client: MongoClient) -> Collection:
    """A Collection with both a Vector and a Full-text Search Index"""
    if COLLECTION_NAME not in client[DB_NAME].list_collection_names():
        clxn = client[DB_NAME].create_collection(COLLECTION_NAME)
    else:
        clxn = client[DB_NAME][COLLECTION_NAME]

    clxn.delete_many({})

    if all([INDEX_NAME != ix["name"] for ix in clxn.list_search_indexes()]):
        index.create_vector_search_index(
            collection=clxn,
            index_name=INDEX_NAME,
            dimensions=DIMENSIONS,
            path="embedding",
            similarity="cosine",
            filters=None,
            wait_until_complete=TIMEOUT,
        )

    return clxn


@pytest.mark.skipif(
    not os.environ.get("OPENAI_API_KEY") and "AZURE_OPENAI_ENDPOINT" not in os.environ,
    reason="Requires OpenAI for chat responses.",
)
def test_chain(
    collection: Collection,
    embedding: Embeddings,
) -> None:
    """Demonstrate usage of MongoDBAtlasVectorSearch in a realistic chain

    Follows example in the docs: https://python.langchain.com/docs/how_to/hybrid/

    Requires OpenAI_API_KEY for embedding and chat model.
    Requires INDEX_NAME to have been set up on MONGODB_URI
    """

    vectorstore = PatchedMongoDBAtlasVectorSearch(
        collection=collection,
        embedding=embedding,
        index_name=INDEX_NAME,
        text_key="page_content",
    )

    texts = [
        "In 2023, I visited Paris",
        "In 2022, I visited New York",
        "In 2021, I visited New Orleans",
        "In 2019, I visited San Francisco",
        "In 2020, I visited Vancouver",
    ]
    vectorstore.add_texts(texts)

    query = "In the United States, what city did I visit last?"
    # One can do vector search on the vector store, using its various search types.
    k = len(texts)

    store_output = list(vectorstore.similarity_search(query=query, k=k))
    assert len(store_output) == k
    assert isinstance(store_output[0], Document)

    # Unfortunately, the VectorStore output cannot be given to a Chat Model
    # If we wish Chat Model to answer based on our own data,
    # we have to give it the right things to work with.
    # The way that Langchain does this is by piping results along in
    # a Chain: https://python.langchain.com/v0.1/docs/modules/chains/

    # Now, we can turn our VectorStore into something Runnable in a Chain
    # by turning it into a Retriever.
    # For the simple VectorSearch Retriever, we can do this like so.

    retriever = vectorstore.as_retriever(search_kwargs=dict(k=k))

    # This does not do much other than expose our search function
    # as an invoke() method with a a certain API, a Runnable.
    retriever_output = retriever.invoke(query)
    assert len(retriever_output) == len(texts)
    assert retriever_output[0].page_content == store_output[0].page_content

    # To get a natural language response to our question,
    # we need ChatOpenAI, a template to better frame the question as a prompt,
    # and a parser to send the output to a string.
    # Together, these become our Chain!
    # Here goes:

    template = """Answer the question based only on the following context.
     Answer in as few words as possible.
     {context}
     Question: {question}
     """
    prompt = ChatPromptTemplate.from_template(template)

    if "AZURE_OPENAI_ENDPOINT" in os.environ:
        model: BaseChatOpenAI = AzureChatOpenAI(model="o4-mini", seed=12345)
    else:
        model = ChatOpenAI(seed=12345)

    chain = (
        {"context": retriever, "question": RunnablePassthrough()}  # type: ignore
        | prompt
        | model
        | StrOutputParser()
    )

    answer = chain.invoke("What city did I visit last?")

    assert "Paris" in answer



import os
import warnings
from typing import Generator, List

import pytest
from langchain_community.document_loaders import PyPDFLoader
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_ollama.embeddings import OllamaEmbeddings
from langchain_openai import AzureOpenAIEmbeddings, OpenAIEmbeddings
from pymongo import MongoClient

from ..utils import CONNECTION_STRING


@pytest.fixture(scope="session")
def technical_report_pages() -> List[Document]:
    """Returns a Document for each of the 100 pages of a GPT-4 Technical Report"""
    loader = PyPDFLoader("https://arxiv.org/pdf/2303.08774.pdf")
    with warnings.catch_warnings():
        # Ignore warnings raised by base class.
        warnings.simplefilter("ignore", ResourceWarning)
        pages = loader.load()
    return pages


@pytest.fixture(scope="session")
def client() -> Generator[MongoClient, None, None]:
    client = MongoClient(CONNECTION_STRING)
    yield client
    client.close()


@pytest.fixture(scope="session")
def embedding() -> Embeddings:
    if os.environ.get("OPENAI_API_KEY"):
        return OpenAIEmbeddings(
            openai_api_key=os.environ["OPENAI_API_KEY"],  # type: ignore # noqa
            model="text-embedding-3-small",
        )
    if os.environ.get("AZURE_OPENAI_ENDPOINT"):
        return AzureOpenAIEmbeddings(model="text-embedding-3-small")

    return OllamaEmbeddings(model="all-minilm:l6-v2")


@pytest.fixture(scope="session")
def dimensions() -> int:
    if os.environ.get("OPENAI_API_KEY") or os.environ.get("AZURE_OPENAI_ENDPOINT"):
        return 1536
    return 384



"""Test MongoDBAtlasVectorSearch.from_documents."""

from __future__ import annotations

from typing import Dict, Generator, List

import pytest  # type: ignore[import-not-found]
from langchain_core.embeddings import Embeddings
from pymongo import MongoClient
from pymongo.collection import Collection

from langchain_mongodb import MongoDBAtlasVectorSearch
from langchain_mongodb.index import (
    create_vector_search_index,
)

from ..utils import DB_NAME, ConsistentFakeEmbeddings, PatchedMongoDBAtlasVectorSearch

COLLECTION_NAME = "langchain_test_from_texts"
INDEX_NAME = "langchain-test-index-from-texts"
DIMENSIONS = 5


@pytest.fixture(scope="module")
def collection(client: MongoClient) -> Collection:
    if COLLECTION_NAME not in client[DB_NAME].list_collection_names():
        clxn = client[DB_NAME].create_collection(COLLECTION_NAME)
    else:
        clxn = client[DB_NAME][COLLECTION_NAME]

    clxn.delete_many({})

    if not any([INDEX_NAME == ix["name"] for ix in clxn.list_search_indexes()]):
        create_vector_search_index(
            collection=clxn,
            index_name=INDEX_NAME,
            dimensions=DIMENSIONS,
            path="embedding",
            filters=["c"],
            similarity="cosine",
            wait_until_complete=60,
        )

    return clxn


@pytest.fixture(scope="module")
def texts() -> List[str]:
    return [
        "Dogs are tough.",
        "Cats have fluff.",
        "What is a sandwich?",
        "That fence is purple.",
    ]


@pytest.fixture(scope="module")
def metadatas() -> List[Dict]:
    return [{"a": 1}, {"b": 1}, {"c": 1}, {"d": 1, "e": 2}]


@pytest.fixture(scope="module")
def embeddings() -> Embeddings:
    return ConsistentFakeEmbeddings(DIMENSIONS)


@pytest.fixture(scope="module")
def vectorstore(
    collection: Collection,
    texts: List[str],
    embeddings: Embeddings,
    metadatas: List[dict],
) -> Generator[MongoDBAtlasVectorSearch]:
    """VectorStore created with a few documents and a trivial embedding model.

    Note: PatchedMongoDBAtlasVectorSearch is MongoDBAtlasVectorSearch in all
    but one important feature. It waits until all documents are fully indexed
    before returning control to the caller.
    """
    vectorstore_from_texts = PatchedMongoDBAtlasVectorSearch.from_texts(
        texts=texts,
        embedding=embeddings,
        metadatas=metadatas,
        collection=collection,
        index_name=INDEX_NAME,
    )
    yield vectorstore_from_texts

    vectorstore_from_texts.collection.delete_many({})


def test_search_with_metadatas_and_pre_filter(
    vectorstore: PatchedMongoDBAtlasVectorSearch, metadatas: List[Dict]
) -> None:
    # Confirm the presence of metadata in output
    output = vectorstore.similarity_search("Sandwich", k=1)
    assert len(output) == 1
    metakeys = [list(d.keys())[0] for d in metadatas]
    assert any([key in output[0].metadata for key in metakeys])


def test_search_filters_all(
    vectorstore: PatchedMongoDBAtlasVectorSearch, metadatas: List[Dict]
) -> None:
    # Test filtering out
    does_not_match_filter = vectorstore.similarity_search(
        "Sandwich", k=1, pre_filter={"c": {"$lte": 0}}
    )
    assert does_not_match_filter == []


def test_search_pre_filter(
    vectorstore: PatchedMongoDBAtlasVectorSearch, metadatas: List[Dict]
) -> None:
    # Test filtering with expected output
    matches_filter = vectorstore.similarity_search(
        "Sandwich", k=3, pre_filter={"c": {"$gt": 0}}
    )
    assert len(matches_filter) == 1


def test_similarity_search_by_vector(
    vectorstore: PatchedMongoDBAtlasVectorSearch,
    embeddings: Embeddings,
    texts: List[str],
) -> None:
    # Test similarity_search_by_vector method
    # First, embed a query text to get a vector
    query_text = "Sandwich"
    query_vector = embeddings.embed_query(query_text)

    # Perform search by vector
    output = vectorstore.similarity_search_by_vector(query_vector, k=2)

    # Should return results
    assert len(output) == 2
    # Results should be Document objects
    assert all(hasattr(doc, "page_content") for doc in output)
    assert all(hasattr(doc, "metadata") for doc in output)


def test_similarity_search_by_vector_with_filter(
    vectorstore: PatchedMongoDBAtlasVectorSearch,
    embeddings: Embeddings,
) -> None:
    # Test similarity_search_by_vector with pre_filter
    query_text = "Sandwich"
    query_vector = embeddings.embed_query(query_text)

    # Search with filter
    filtered_output = vectorstore.similarity_search_by_vector(
        query_vector, k=3, pre_filter={"c": {"$gt": 0}}
    )

    # Should only return documents matching the filter
    assert len(filtered_output) == 1
    assert "c" in filtered_output[0].metadata



from typing import Dict, List

from langchain_core.documents import Document
from pymongo import MongoClient

from langchain_mongodb.loaders import MongoDBLoader

from ..utils import DB_NAME

COLLECTION_NAME = "langchain_test_loader"


def raw_docs() -> List[Dict]:
    return [
        {"_id": "1", "address": {"building": "1", "room": "1"}},
        {"_id": "2", "address": {"building": "2", "room": "2"}},
        {"_id": "3", "address": {"building": "3", "room": "2"}},
    ]


def expected_documents() -> List[Document]:
    return [
        Document(
            page_content="2 2",
            metadata={"_id": "2", "database": DB_NAME, "collection": COLLECTION_NAME},
        ),
        Document(
            page_content="3 2",
            metadata={"_id": "3", "database": DB_NAME, "collection": COLLECTION_NAME},
        ),
    ]


async def test_load_with_filters(client: MongoClient) -> None:
    filter_criteria = {"address.room": {"$eq": "2"}}
    field_names = ["address.building", "address.room"]
    metadata_names = ["_id"]
    include_db_collection_in_metadata = True

    collection = client[DB_NAME][COLLECTION_NAME]
    collection.delete_many({})
    collection.insert_many(raw_docs())

    loader = MongoDBLoader(
        collection,
        filter_criteria=filter_criteria,
        field_names=field_names,
        metadata_names=metadata_names,
        include_db_collection_in_metadata=include_db_collection_in_metadata,
    )
    documents = await loader.aload()

    assert documents == expected_documents()



"""
The following contains the JSON Schema for the Entities in the Collection representing the Knowledge Graph.
If validate is set to True, the schema is enforced upon insert and update.
See `$jsonSchema <https://www.mongodb.com/docs/manual/reference/operator/query/jsonSchema/>`_

The following defines the default entity_schema.
It allows all possible values of "type" and "relationship".

If allowed_entity_types: List[str] is given to MongoDBGraphStore's constructor,
then `self._schema["properties"]["type"]["enum"] = allowed_entity_types` is added.

If allowed_relationship_types: List[str] is given to MongoDBGraphStore's constructor,
additionalProperties is set to False, and relationship schema is provided for each key.
"""

entity_schema = {
    "bsonType": "object",
    "required": ["_id", "type"],
    "properties": {
        "_id": {
            "bsonType": "string",
            "description": "Unique identifier for the entity",
        },
        "type": {
            "bsonType": "string",
            "description": "Type of the entity (e.g., 'Person', 'Organization')",
            # Note: When constrained, predefined types are added. For example:
            # "enum": ["Person", "Organization", "Location", "Event"],
        },
        "attributes": {
            "bsonType": "object",
            "description": "Key-value pairs describing the entity",
            "additionalProperties": {
                "bsonType": "array",
                "items": {"bsonType": "string"},  # Enforce array of strings
            },
        },
        "relationships": {
            "bsonType": "object",
            "description": "Key-value pairs of relationships",
            "required": ["target_ids"],
            "properties": {
                "target_ids": {
                    "bsonType": "array",
                    "description": "name/_id values of the target entities",
                    "items": {"bsonType": "string"},
                },
                "types": {
                    "bsonType": "array",
                    "description": "An array of relationships to corresponding target_ids (in same array position).",
                    "items": {"bsonType": "string"},
                    # Note: When constrained, predefined types are added. For example:
                    # "enum": ["used_in", "owns", "written_by", "located_in"],  # Predefined types
                },
                "attributes": {
                    "bsonType": "array",
                    "description": "An array of attributes describing the relationships to corresponding target_ids (in same array position). Each element is an object containing key-value pairs, where values are arrays of strings.",
                    "items": {
                        "bsonType": "object",
                        "additionalProperties": {
                            "bsonType": "array",
                            "items": {"bsonType": "string"},
                        },
                    },
                },
            },
            "additionalProperties": False,
        },
    },
}



"""Test max_marginal_relevance_search."""

from __future__ import annotations

import pytest  # type: ignore[import-not-found]
from langchain_core.embeddings import Embeddings
from pymongo import MongoClient
from pymongo.collection import Collection

from langchain_mongodb.index import (
    create_vector_search_index,
)

from ..utils import DB_NAME, ConsistentFakeEmbeddings, PatchedMongoDBAtlasVectorSearch

COLLECTION_NAME = "langchain_test_vectorstores"
INDEX_NAME = "langchain-test-index-vectorstores"
DIMENSIONS = 5


@pytest.fixture()
def collection(client: MongoClient) -> Collection:
    if COLLECTION_NAME not in client[DB_NAME].list_collection_names():
        clxn = client[DB_NAME].create_collection(COLLECTION_NAME)
    else:
        clxn = client[DB_NAME][COLLECTION_NAME]

    clxn.delete_many({})

    if not any([INDEX_NAME == ix["name"] for ix in clxn.list_search_indexes()]):
        create_vector_search_index(
            collection=clxn,
            index_name=INDEX_NAME,
            dimensions=5,
            path="embedding",
            filters=["c"],
            similarity="cosine",
            wait_until_complete=60,
        )

    return clxn


@pytest.fixture
def embeddings() -> Embeddings:
    return ConsistentFakeEmbeddings(DIMENSIONS)


def test_mmr(embeddings: Embeddings, collection: Collection) -> None:
    texts = ["foo", "foo", "fou", "foy"]
    collection.delete_many({})
    vectorstore = PatchedMongoDBAtlasVectorSearch.from_texts(
        texts,
        embedding=embeddings,
        collection=collection,
        index_name=INDEX_NAME,
    )
    query = "foo"
    output = vectorstore.max_marginal_relevance_search(query, k=10, lambda_mult=0.1)
    assert len(output) == len(texts)
    assert output[0].page_content == "foo"
    assert output[1].page_content != "foo"



from __future__ import annotations

import os
from copy import deepcopy
from time import monotonic, sleep
from typing import Any, Dict, Generator, Iterable, List, Mapping, Optional, Union, cast

from bson import ObjectId
from langchain_core.callbacks.manager import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain_core.embeddings import Embeddings
from langchain_core.language_models.chat_models import BaseChatModel, SimpleChatModel
from langchain_core.language_models.llms import LLM
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
)
from langchain_core.outputs import ChatGeneration, ChatResult
from langchain_ollama import ChatOllama
from langchain_openai import AzureChatOpenAI, ChatOpenAI
from pydantic import model_validator
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.driver_info import DriverInfo
from pymongo.operations import SearchIndexModel
from pymongo.results import BulkWriteResult, DeleteResult, InsertManyResult

from langchain_mongodb import MongoDBAtlasVectorSearch
from langchain_mongodb.agent_toolkit.database import MongoDBDatabase
from langchain_mongodb.cache import MongoDBAtlasSemanticCache

TIMEOUT = 120
INTERVAL = 0.5
CONNECTION_STRING = os.environ.get("MONGODB_URI", "")


DB_NAME = "langchain_test_db"


def create_database() -> MongoDBDatabase:
    client = MongoClient(CONNECTION_STRING)
    coll = client[DB_NAME]["test"]
    coll.delete_many({})
    coll.insert_one({})
    return MongoDBDatabase(client, DB_NAME)


def create_llm() -> BaseChatModel:
    if os.environ.get("AZURE_OPENAI_ENDPOINT"):
        return AzureChatOpenAI(model="o4-mini", timeout=60, cache=False, seed=12345)
    if os.environ.get("OPENAI_API_KEY"):
        return ChatOpenAI(model="gpt-4o-mini", timeout=60, cache=False, seed=12345)
    return ChatOllama(model="llama3:8b", cache=False, seed=12345)


class PatchedMongoDBAtlasVectorSearch(MongoDBAtlasVectorSearch):
    def bulk_embed_and_insert_texts(
        self,
        texts: Union[List[str], Iterable[str]],
        metadatas: Union[List[dict], Generator[dict, Any, Any]],
        ids: Optional[List[str]] = None,
    ) -> List:
        """Patched insert_texts that waits for data to be indexed before returning"""
        ids_inserted = super().bulk_embed_and_insert_texts(texts, metadatas, ids)
        n_docs = self.collection.count_documents({})
        start = monotonic()
        while monotonic() - start <= TIMEOUT:
            if (
                len(self.similarity_search("sandwich", k=n_docs, oversampling_factor=1))
                == n_docs
            ):
                return ids_inserted
            else:
                sleep(INTERVAL)
        raise TimeoutError(f"Failed to embed, insert, and index texts in {TIMEOUT}s.")

    def _similarity_search_with_score(self, query_vector, **kwargs):
        # Remove the _ids for testing purposes.
        docs = super()._similarity_search_with_score(query_vector, **kwargs)
        for doc, _ in docs:
            del doc.metadata["_id"]
        return docs

    def delete(self, ids: Optional[List[str]] = None, **kwargs: Any) -> Optional[bool]:
        ret = super().delete(ids, **kwargs)
        n_docs = self.collection.count_documents({})
        start = monotonic()
        while monotonic() - start <= TIMEOUT:
            if (
                len(
                    self.similarity_search(
                        "sandwich", k=max(n_docs, 1), oversampling_factor=1
                    )
                )
                == n_docs
            ):
                return ret
            else:
                sleep(INTERVAL)
        raise TimeoutError(f"Failed to embed, insert, and index texts in {TIMEOUT}s.")


class ConsistentFakeEmbeddings(Embeddings):
    """Fake embeddings functionality for testing."""

    def __init__(self, dimensionality: int = 10) -> None:
        self.known_texts: List[str] = []
        self.dimensionality = dimensionality

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Return consistent embeddings for each text seen so far."""
        out_vectors = []
        for text in texts:
            if text not in self.known_texts:
                self.known_texts.append(text)
            vector = [1.0] * (self.dimensionality - 1) + [
                float(self.known_texts.index(text))
            ]
            out_vectors.append(vector)
        return out_vectors

    def embed_query(self, text: str) -> List[float]:
        """Return consistent embeddings for the text, if seen before, or a constant
        one if the text is unknown."""
        return self.embed_documents([text])[0]

    async def aembed_documents(self, texts: List[str]) -> List[List[float]]:
        return self.embed_documents(texts)

    async def aembed_query(self, text: str) -> List[float]:
        return self.embed_query(text)


class FakeChatModel(SimpleChatModel):
    """Fake Chat Model wrapper for testing purposes."""

    def _call(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        return "fake response"

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        output_str = "fake response"
        message = AIMessage(content=output_str)
        generation = ChatGeneration(message=message)
        return ChatResult(generations=[generation])

    @property
    def _llm_type(self) -> str:
        return "fake-chat-model"

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        return {"key": "fake"}


class FakeLLM(LLM):
    """Fake LLM wrapper for testing purposes."""

    queries: Optional[Mapping] = None
    sequential_responses: Optional[bool] = False
    response_index: int = 0

    @model_validator(mode="before")
    @classmethod
    def check_queries_required(cls, values: dict) -> dict:
        if values.get("sequential_response") and not values.get("queries"):
            raise ValueError(
                "queries is required when sequential_response is set to True"
            )
        return values

    def get_num_tokens(self, text: str) -> int:
        """Return number of tokens."""
        return len(text.split())

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "fake"

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        if self.sequential_responses:
            return self._get_next_response_in_sequence
        if self.queries is not None:
            return self.queries[prompt]
        if stop is None:
            return "foo"
        else:
            return "bar"

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        return {}

    @property
    def _get_next_response_in_sequence(self) -> str:
        queries = cast(Mapping, self.queries)
        response = queries[list(queries.keys())[self.response_index]]
        self.response_index = self.response_index + 1
        return response


class MockClient:
    is_closed = False

    def __getitem__(self, key: str) -> Any:
        return MockDatabase(self)

    def close(self):
        self.is_closed = True

    def append_metadata(self, metadata: DriverInfo) -> None:
        pass


class MockDatabase:
    name = "test"

    def __init__(self, client=None):
        self.client = client or MockClient()

    def list_collection_names(self, authorizedCollections: bool = True) -> list[str]:
        return ["test"]

    def __getitem__(self, key: str) -> Any:
        return MockCollection(self)


class MockCollection(Collection):
    """Mocked Mongo Collection"""

    _aggregate_result: List[Any]
    _insert_result: Optional[InsertManyResult]
    _data: List[Any]
    _simulate_cache_aggregation_query: bool

    def __init__(self, database: MockDatabase | None = None) -> None:
        self._data = []
        self._name = "test"
        self.is_closed = False
        self._aggregate_result = []
        self._insert_result = None
        self._search_indexes = []
        self._simulate_cache_aggregation_query = False
        self._database = database or MockDatabase()  # type:ignore[assignment]

    @property
    def database(self):
        return self._database

    def close(self):
        self.is_closed = True

    def list_search_indexes(self, name=None, session=None, comment=None, **kwargs):
        return [
            dict(name=idx.document["name"], status="READY")
            for idx in self._search_indexes
        ]

    def create_search_index(self, model, session=None, comment=None, **kwargs):
        if not isinstance(model, SearchIndexModel):
            model = SearchIndexModel(model, name=f"test{len(self._search_indexes)}")
        self._search_indexes.append(model)

    def delete_many(self, *args, **kwargs) -> DeleteResult:  # type: ignore
        old_len = len(self._data)
        self._data = []
        return DeleteResult({"n": old_len}, acknowledged=True)

    def insert_many(self, to_insert: List[Any], *args, **kwargs) -> InsertManyResult:  # type: ignore
        mongodb_inserts = [
            {"_id": ObjectId(), "score": 1, **insert} for insert in to_insert
        ]
        self._data.extend(mongodb_inserts)
        return self._insert_result or InsertManyResult(
            [k["_id"] for k in mongodb_inserts], acknowledged=True
        )

    def insert_one(self, to_insert: Any, *args, **kwargs) -> Any:  # type: ignore
        return self.insert_many([to_insert])

    def find_one(self, find_query: Dict[str, Any]) -> Optional[Dict[str, Any]]:  # type: ignore
        find = self.find(find_query) or [None]  # type: ignore
        return find[0]

    def find(self, find_query: Dict[str, Any]) -> Optional[List[Dict[str, Any]]]:  # type: ignore
        def _is_match(item: Dict[str, Any]) -> bool:
            for key, match_val in find_query.items():
                if item.get(key) != match_val:
                    return False
            return True

        return [document for document in self._data if _is_match(document)]

    def update_one(  # type: ignore
        self,
        find_query: Dict[str, Any],
        options: Dict[str, Any],
        *args: Any,
        upsert=True,
        **kwargs: Any,
    ) -> None:  # type: ignore
        result = self.find_one(find_query)
        set_options = options.get("$set", {})

        if result:
            result.update(set_options)
        elif upsert:
            self._data.append({**find_query, **set_options})

    def _execute_cache_aggregation_query(self, *args, **kwargs) -> List[Dict[str, Any]]:  # type: ignore
        """Helper function only to be used for MongoDBAtlasSemanticCache Testing

        Returns:
            List[Dict[str, Any]]: Aggregation query result
        """
        pipeline: List[Dict[str, Any]] = args[0]
        params = pipeline[0]["$vectorSearch"]
        embedding = params["queryVector"]
        # Assumes MongoDBAtlasSemanticCache.LLM == "llm_string"
        llm_string = params["filter"][MongoDBAtlasSemanticCache.LLM]["$eq"]

        acc = []
        for document in self._data:
            if (
                document.get("embedding") == embedding
                and document.get(MongoDBAtlasSemanticCache.LLM) == llm_string
            ):
                acc.append(document)
        return acc

    def bulk_write(self, requests, **kwargs):
        upserted = []
        for ind, request in enumerate(requests):
            doc = request._doc
            doc["score"] = "foo"
            self._data.append(doc)
            upserted.append(dict(index=ind, _id=doc["_id"]))
        return BulkWriteResult(dict(upserted=upserted), True)

    def aggregate(self, *args, **kwargs) -> List[Any]:  # type: ignore
        if self._simulate_cache_aggregation_query:
            return deepcopy(self._execute_cache_aggregation_query(*args, **kwargs))
        return deepcopy(self._aggregate_result)

    def count_documents(self, *args, **kwargs) -> int:  # type: ignore
        return len(self._data)

    def __repr__(self) -> str:
        return "MockCollection"



import json
import warnings

from langchain_classic.memory import (
    ConversationBufferMemory,
)
from langchain_core.messages import message_to_dict

from langchain_mongodb.chat_message_histories import MongoDBChatMessageHistory

from ..utils import CONNECTION_STRING, DB_NAME

COLLECTION = "langchain_test_chat"


def test_memory_with_message_store() -> None:
    """Test the memory with a message store."""
    # setup MongoDB as a message store
    message_history = MongoDBChatMessageHistory(
        connection_string=CONNECTION_STRING,
        session_id="test-session",
        database_name=DB_NAME,
        collection_name=COLLECTION,
    )
    with warnings.catch_warnings():
        # Ignore warnings raised by base class.
        warnings.simplefilter("ignore", DeprecationWarning)
        memory = ConversationBufferMemory(
            memory_key="baz", chat_memory=message_history, return_messages=True
        )

    # add some messages
    memory.chat_memory.add_ai_message("This is me, the AI")
    memory.chat_memory.add_user_message("This is me, the human")

    # get the message history from the memory store and turn it into a json
    messages = memory.chat_memory.messages
    messages_json = json.dumps([message_to_dict(msg) for msg in messages])

    assert "This is me, the AI" in messages_json
    assert "This is me, the human" in messages_json

    # remove the record from MongoDB, so the next test run won't pick it up
    memory.chat_memory.clear()

    assert memory.chat_memory.messages == []
    memory.chat_memory.close()  # type:ignore[attr-defined]



"""Wrapper around a MongoDB database."""

from __future__ import annotations

import json
import re
from datetime import date, datetime, timezone
from typing import Any, Dict, Iterable, List, Optional, Union

from bson import ObjectId
from bson.binary import Binary
from bson.decimal128 import Decimal128
from bson.json_util import dumps
from pymongo import MongoClient
from pymongo.cursor import Cursor
from pymongo.errors import PyMongoError

from langchain_mongodb.utils import DRIVER_METADATA, _append_client_metadata

NUM_DOCUMENTS_TO_SAMPLE = 4
MAX_STRING_LENGTH_OF_SAMPLE_DOCUMENT_VALUE = 20

_BSON_LOOKUP = {
    str: "String",
    int: "Number",
    float: "Number",
    bool: "Boolean",
    ObjectId: "ObjectId",
    date: "Date",
    datetime: "Timestamp",
    None: "Null",
    Decimal128: "Decimal128",
    Binary: "Binary",
}


class MongoDBDatabase:
    """Wrapper around a MongoDB database."""

    def __init__(
        self,
        client: MongoClient,
        database: str,
        schema: Optional[str] = None,
        ignore_collections: Optional[List[str]] = None,
        include_collections: Optional[List[str]] = None,
        sample_docs_in_collection_info: int = 3,
        indexes_in_collection_info: bool = False,
    ):
        """Create a MongoDBDatabase from client and database name."""
        self._client = client
        self._db = client[database]
        self._schema = schema
        if include_collections and ignore_collections:
            raise ValueError(
                "Cannot specify both include_collections and ignore_collections"
            )
        self._include_colls = set(include_collections or [])
        self._ignore_colls = set(ignore_collections or [])
        self._all_colls = set(
            self._db.list_collection_names(authorizedCollections=True)
        )

        self._sample_docs_in_coll_info = sample_docs_in_collection_info
        self._indexes_in_coll_info = indexes_in_collection_info

        _append_client_metadata(self._client)

    @classmethod
    def from_connection_string(
        cls,
        connection_string: str,
        database: Optional[str] = None,
        **kwargs: Any,
    ) -> MongoDBDatabase:
        """Construct a MongoDBDatabase from URI."""
        client: MongoClient[dict[str, Any]] = MongoClient(
            connection_string,
            driver=DRIVER_METADATA,
        )
        database = database or client.get_default_database().name
        return cls(client, database, **kwargs)

    def close(self) -> None:
        """Close the resources used by the MongoDBDatabase."""
        self._client.close()

    def get_usable_collection_names(self) -> Iterable[str]:
        """Get names of collections available."""
        if self._include_colls:
            return sorted(self._include_colls)

        return sorted(self._all_colls - self._ignore_colls)

    @property
    def collection_info(self) -> str:
        """Information about all collections in the database."""
        return self.get_collection_info()

    def get_collection_info(self, collection_names: Optional[List[str]] = None) -> str:
        """Get information about specified collections.

        Follows best practices as specified in: Rajkumar et al, 2022
        (https://arxiv.org/abs/2204.00498)

        If `sample_rows_in_collection_info`, the specified number of sample rows will be
        appended to each collection description. This can increase performance as
        demonstrated in the paper.
        """
        all_coll_names = self.get_usable_collection_names()
        if collection_names is not None:
            missing_collections = set(collection_names).difference(all_coll_names)
            if missing_collections:
                raise ValueError(
                    f"collection_names {missing_collections} not found in database"
                )
            all_coll_names = collection_names

        colls = []
        for coll in all_coll_names:
            # add schema
            schema = self._get_collection_schema(coll)
            coll_info = f"Database name: {self._db.name}\n"
            coll_info += f"Collection name: {coll}\n"
            coll_info += f"Schema from a sample of documents from the collection:\n{schema.rstrip()}"
            has_extra_info = (
                self._indexes_in_coll_info or self._sample_docs_in_coll_info
            )
            if has_extra_info:
                coll_info += "\n\n/*"
            if self._indexes_in_coll_info:
                coll_info += f"\n{self._get_collection_indexes(coll)}\n"
            if self._sample_docs_in_coll_info:
                coll_info += f"\n{self._get_sample_docs(coll)}\n"
            if has_extra_info:
                coll_info += "*/"
            colls.append(coll_info)
        colls.sort()
        final_str = "\n\n".join(colls)
        return final_str

    def _get_collection_schema(self, collection: str) -> str:
        coll = self._db[collection]
        doc = coll.find_one({}) or dict()
        return "\n".join(self._parse_doc(doc, ""))

    def _parse_doc(self, doc: dict[str, Any], prefix: str) -> list[str]:
        sub_schema = []
        for key, value in doc.items():
            if prefix:
                full_key = f"{prefix}.{key}"
            else:
                full_key = key
            if isinstance(value, dict):
                sub_schema.extend(self._parse_doc(value, full_key))
            elif isinstance(value, list):
                if not len(value):
                    sub_schema.append(f"{full_key}: Array")
                elif isinstance(value[0], dict):
                    sub_schema.extend(self._parse_doc(value[0], f"{full_key}[]"))
                else:
                    if type(value[0]) in _BSON_LOOKUP:
                        type_name = _BSON_LOOKUP[type(value[0])]
                        sub_schema.append(f"{full_key}: Array<{type_name}>")
                    else:
                        sub_schema.append(f"{full_key}: Array")
            elif type(value) in _BSON_LOOKUP:
                type_name = _BSON_LOOKUP[type(value)]
                sub_schema.append(f"{full_key}: {type_name}")
        if not sub_schema:
            sub_schema.append(f"{prefix}: Document")
        return sub_schema

    def _get_collection_indexes(self, collection: str) -> str:
        coll = self._db[collection]
        indexes = list(coll.list_indexes())
        if not indexes:
            return ""
        return f"Collection Indexes:\n{json.dumps(indexes, indent=2)}"

    def _get_sample_docs(self, collection: str) -> str:
        col = self._db[collection]
        docs = list(col.find({}, limit=self._sample_docs_in_coll_info))
        for doc in docs:
            self._elide_doc(doc)
        return (
            f"{self._sample_docs_in_coll_info} documents from {collection} collection:\n"
            f"{dumps(docs, indent=2)}"
        )

    def _elide_doc(self, doc: dict[str, Any]) -> None:
        for key, value in doc.items():
            if isinstance(value, dict):
                self._elide_doc(value)
            elif isinstance(value, list):
                items = []
                for item in value:
                    if isinstance(item, dict):
                        self._elide_doc(item)
                    elif (
                        isinstance(item, str)
                        and len(item) > MAX_STRING_LENGTH_OF_SAMPLE_DOCUMENT_VALUE
                    ):
                        item = item[: MAX_STRING_LENGTH_OF_SAMPLE_DOCUMENT_VALUE + 1]
                    items.append(item)
                doc[key] = items
            elif (
                isinstance(value, str)
                and len(value) > MAX_STRING_LENGTH_OF_SAMPLE_DOCUMENT_VALUE
            ):
                doc[key] = value[: MAX_STRING_LENGTH_OF_SAMPLE_DOCUMENT_VALUE + 1]

    def _parse_command(self, command: str) -> Any:
        """
        Extracts and parses the aggregation pipeline from a JavaScript-style MongoDB command.
        Handles ObjectId(), ISODate(), new Date() and converts them into Python constructs.
        """
        command = re.sub(r"\s+", " ", command.strip())
        if command.endswith("]"):
            command += ")"

        try:
            agg_str = command.split(".aggregate(", 1)[1].rsplit(")", 1)[0]
        except Exception as e:
            raise ValueError(f"Could not extract aggregation pipeline: {e}") from e

        # Convert JavaScript-style constructs to Python syntax
        agg_str = self._convert_mongo_js_to_python(agg_str)

        try:
            eval_globals = {
                "ObjectId": ObjectId,
                "datetime": datetime,
                "timezone": timezone,
            }
            agg_pipeline = eval(agg_str, eval_globals)
            if not isinstance(agg_pipeline, list):
                raise ValueError("Aggregation pipeline must be a list.")
            return agg_pipeline
        except Exception as e:
            raise ValueError(f"Failed to parse aggregation pipeline: {e}") from e

    def run(self, command: str) -> Union[str, Cursor]:
        """Execute a MongoDB aggregation command and return a string representing the results.

        If the statement returns documents, a string of the results is returned.
        If the statement returns no documents, an empty string is returned.

        The command MUST be of the form: `db.collectionName.aggregate(...)`.
        """
        if not command.startswith("db."):
            raise ValueError(f"Cannot run command {command}")

        try:
            col_name = command.split(".")[1]
        except IndexError as e:
            raise ValueError(
                "Invalid command format. Could not extract collection name."
            ) from e

        if col_name not in self.get_usable_collection_names():
            raise ValueError(f"Collection {col_name} does not exist!")

        if ".aggregate(" not in command:
            raise ValueError("Only aggregate(...) queries are currently supported.")

        # Parse pipeline using helper
        agg_pipeline = self._parse_command(command)

        try:
            coll = self._db[col_name]
            result = coll.aggregate(agg_pipeline)
            return dumps(list(result), indent=2)
        except Exception as e:
            raise ValueError(f"Error executing aggregation: {e}") from e

    def get_collection_info_no_throw(
        self, collection_names: Optional[List[str]] = None
    ) -> str:
        """Get information about specified collections.

        Follows best practices as specified in: Rajkumar et al, 2022
        (https://arxiv.org/abs/2204.00498)

        If `sample_rows_in_collection_info`, the specified number of sample rows will be
        appended to each collection description. This can increase performance as
        demonstrated in the paper.
        """
        try:
            return self.get_collection_info(collection_names)
        except ValueError as e:
            """Format the error message"""
            raise e
            return f"Error: {e}"

    def run_no_throw(self, command: str) -> Union[str, Cursor]:
        """Execute a MongoDB command and return a string representing the results.

        If the statement returns rows, a string of the results is returned.
        If the statement returns no rows, an empty string is returned.

        If the statement throws an error, the error message is returned.
        """
        try:
            return self.run(command)
        except PyMongoError as e:
            """Format the error message"""
            return f"Error: {e}"

    def get_context(self) -> Dict[str, Any]:
        """Return db context that you may want in agent prompt."""
        collection_names = list(self.get_usable_collection_names())
        collection_info = self.get_collection_info_no_throw()
        return {
            "collection_info": collection_info,
            "collection_names": ", ".join(collection_names),
        }

    def _convert_mongo_js_to_python(self, code: str) -> str:
        """Convert JavaScript-style MongoDB syntax into Python-safe code."""

        def _handle_iso_date(match: Any) -> str:
            date_str = match.group(1)
            if not date_str:
                raise ValueError("ISODate must contain a date string.")
            dt = datetime.fromisoformat(date_str.replace("Z", "+00:00"))
            return f"datetime({dt.year}, {dt.month}, {dt.day}, {dt.hour}, {dt.minute}, {dt.second}, tzinfo=timezone.utc)"

        def _handle_new_date(match: Any) -> str:
            date_str = match.group(1)
            if not date_str:
                raise ValueError(
                    "new Date() without arguments is not allowed. Please pass an explicit date string."
                )
            dt = datetime.fromisoformat(date_str.replace("Z", "+00:00"))
            return f"datetime({dt.year}, {dt.month}, {dt.day}, {dt.hour}, {dt.minute}, {dt.second}, tzinfo=timezone.utc)"

        def _handle_object_id(match: Any) -> str:
            oid_str = match.group(1)
            if not oid_str:
                raise ValueError("ObjectId must contain a value.")
            return f"ObjectId('{oid_str}')"

        def _handle_id_key(match: Any) -> str:
            return f'"{match.group(1)}"'

        patterns = [
            (r'ISODate\(\s*["\']([^"\']*)["\']\s*\)', _handle_iso_date),
            (r'new\s+Date\(\s*["\']([^"\']*)["\']\s*\)', _handle_new_date),
            (r'ObjectId\(\s*["\']([^"\']*)["\']\s*\)', _handle_object_id),
            (r'ObjectId\(\s*["\']([^"\']*)["\']\s*\)', _handle_object_id),
            (r'(?<!["\'])\b(_id)\b(?!["\'])', _handle_id_key),
        ]

        for pattern, replacer in patterns:
            code = re.sub(pattern, replacer, code)

        return code



"""Tools for interacting with a MongoDB database."""

from __future__ import annotations

from typing import Any, Dict, Optional, Type

from langchain_core.callbacks import (
    CallbackManagerForToolRun,
)
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import PromptTemplate
from langchain_core.tools import BaseTool
from pydantic import BaseModel, ConfigDict, Field, model_validator
from pymongo.cursor import Cursor

from .database import MongoDBDatabase
from .prompt import MONGODB_QUERY_CHECKER


class BaseMongoDBDatabaseTool(BaseModel):
    """Base tool for interacting with a MongoDB database."""

    db: MongoDBDatabase = Field(exclude=True)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )


class _QueryMongoDBDatabaseToolInput(BaseModel):
    query: str = Field(..., description="A detailed and correct MongoDB query.")


class QueryMongoDBDatabaseTool(BaseMongoDBDatabaseTool, BaseTool):  # type: ignore[override, override]
    """Tool for querying a MongoDB database."""

    name: str = "mongodb_query"
    description: str = """
    Execute a MongoDB query against the database and get back the result.
    If the query is not correct, an error message will be returned.
    If an error is returned, rewrite the query, check the query, and try again.
    """
    args_schema: Type[BaseModel] = _QueryMongoDBDatabaseToolInput

    def _run(self, query: str, **kwargs: Any) -> str | Cursor:
        """Execute the query, return the results or an error message."""
        return self.db.run_no_throw(query)


class _InfoMongoDBDatabaseToolInput(BaseModel):
    collection_names: str = Field(
        ...,
        description=(
            "A comma-separated list of the collection names for which to return the schema. "
            "Example input: 'collection1, collection2, collection3'"
        ),
    )


class InfoMongoDBDatabaseTool(BaseMongoDBDatabaseTool, BaseTool):  # type: ignore[override, override]
    """Tool for getting metadata about a MongoDB database."""

    name: str = "mongodb_schema"
    description: str = (
        "Get the schema and sample documents for the specified MongoDB collections."
    )
    args_schema: Type[BaseModel] = _InfoMongoDBDatabaseToolInput

    def _run(
        self,
        collection_names: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        """Get the schema for collections in a comma-separated list."""
        return self.db.get_collection_info_no_throw(
            [t.strip() for t in collection_names.split(",")]
        )


class _ListMongoDBDatabaseToolInput(BaseModel):
    tool_input: str = Field("", description="An empty string")


class ListMongoDBDatabaseTool(BaseMongoDBDatabaseTool, BaseTool):  # type: ignore[override, override]
    """Tool for getting collection names."""

    name: str = "mongodb_list_collections"
    description: str = "Input is an empty string, output is a comma-separated list of collections in the database."
    args_schema: Type[BaseModel] = _ListMongoDBDatabaseToolInput

    def _run(
        self,
        tool_input: str = "",
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        """Get a comma-separated list of collection names."""
        return ", ".join(self.db.get_usable_collection_names())


class _QueryMongoDBCheckerToolInput(BaseModel):
    query: str = Field(..., description="A detailed and MongoDB query to be checked.")


class QueryMongoDBCheckerTool(BaseMongoDBDatabaseTool, BaseTool):  # type: ignore[override, override]
    """Use an LLM to check if a query is correct.
    Adapted from https://www.patterns.app/blog/2023/01/18/crunchbot-sql-analyst-gpt/"""

    template: str = MONGODB_QUERY_CHECKER
    llm: BaseLanguageModel
    prompt: PromptTemplate = Field(init=False)
    name: str = "mongodb_query_checker"
    description: str = """
    Use this tool to double check if your query is correct before executing it.
    Always use this tool before executing a query with mongodb_query!
    """
    args_schema: Type[BaseModel] = _QueryMongoDBCheckerToolInput

    @model_validator(mode="before")
    @classmethod
    def initialize_prompt(cls, values: Dict[str, Any]) -> Any:
        if "prompt" not in values:
            values["prompt"] = PromptTemplate(
                template=MONGODB_QUERY_CHECKER, input_variables=["query"]
            )

        if values["prompt"].input_variables != ["query"]:
            raise ValueError(
                "Prompt for QueryCheckerTool must have input variables ['query']"
            )

        return values

    def _run(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        """Use the LLM to check the query."""
        # TODO: check the query using pymongo first.
        chain = self.prompt | self.llm
        return chain.invoke(query)  # type:ignore[arg-type]



"""Aggregation pipeline components used in Atlas Full-Text, Vector, and Hybrid Search

See the following for more:
    - `Full-Text Search <https://www.mongodb.com/docs/atlas/atlas-search/aggregation-stages/search/#mongodb-pipeline-pipe.-search>`_
    - `MongoDB Operators <https://www.mongodb.com/docs/atlas/atlas-search/operators-and-collectors/#std-label-operators-ref>`_
    - `Vector Search <https://www.mongodb.com/docs/atlas/atlas-vector-search/vector-search-stage/>`_
    - `Filter Example <https://www.mongodb.com/docs/atlas/atlas-vector-search/vector-search-stage/#atlas-vector-search-pre-filter>`_
"""

from typing import Any, Dict, List, Optional, Union

from pymongo_search_utils import (
    combine_pipelines,  # noqa: F401
    final_hybrid_stage,  # noqa: F401
    reciprocal_rank_stage,  # noqa: F401
    vector_search_stage,  # noqa: F401
)


def text_search_stage(
    query: str,
    search_field: Union[str, List[str]],
    index_name: str,
    limit: Optional[int] = None,
    filter: Optional[Dict[str, Any]] = None,
    include_scores: Optional[bool] = True,
    **kwargs: Any,
) -> List[Dict[str, Any]]:  # noqa: E501
    """Full-Text search using Lucene's standard (BM25) analyzer

    Args:
        query: Input text to search for
        search_field: Field in Collection that will be searched
        index_name: Atlas Search Index name
        limit: Maximum number of documents to return. Default of no limit
        filter: Any MQL match expression comparing an indexed field
        include_scores: Scores provide measure of relative relevance

    Returns:
        Dictionary defining the $search stage
    """
    pipeline = [
        {
            "$search": {
                "index": index_name,
                "text": {"query": query, "path": search_field},
            }
        }
    ]
    if filter:
        pipeline.append({"$match": filter})  # type: ignore
    if include_scores:
        pipeline.append({"$set": {"score": {"$meta": "searchScore"}}})
    if limit:
        pipeline.append({"$limit": limit})  # type: ignore

    return pipeline  # type: ignore



"""Various Utility Functions

- Tools for handling bson.ObjectId

The help IDs live as ObjectId in MongoDB and str in Langchain and JSON.


- Tools for the Maximal Marginal Relevance (MMR) reranking

These are duplicated from langchain_community to avoid cross-dependencies.

Functions "maximal_marginal_relevance" and "cosine_similarity"
are duplicated in this utility respectively from modules:

    - "libs/community/langchain_community/vectorstores/utils.py"
    - "libs/community/langchain_community/utils/math.py"
"""

from __future__ import annotations

import logging
from datetime import date, datetime
from importlib.metadata import version
from typing import Any, Dict, List, Union

import numpy as np
from pymongo import MongoClient
from pymongo.driver_info import DriverInfo
from pymongo_search_utils import append_client_metadata

logger = logging.getLogger(__name__)

Matrix = Union[List[List[float]], List[np.ndarray], np.ndarray]

DRIVER_METADATA = DriverInfo(name="Langchain", version=version("langchain-mongodb"))


def _append_client_metadata(client: MongoClient) -> None:
    append_client_metadata(client=client, driver_info=DRIVER_METADATA)


def cosine_similarity(X: Matrix, Y: Matrix) -> np.ndarray:
    """Row-wise cosine similarity between two equal-width matrices."""
    if len(X) == 0 or len(Y) == 0:
        return np.array([])

    X = np.array(X)
    Y = np.array(Y)
    if X.shape[1] != Y.shape[1]:
        raise ValueError(
            f"Number of columns in X and Y must be the same. X has shape {X.shape} "
            f"and Y has shape {Y.shape}."
        )
    try:
        import simsimd as simd

        X = np.array(X, dtype=np.float32)
        Y = np.array(Y, dtype=np.float32)
        Z = 1 - np.array(simd.cdist(X, Y, metric="cosine"))
        return Z
    except ImportError:
        logger.debug(
            "Unable to import simsimd, defaulting to NumPy implementation. If you want "
            "to use simsimd please install with `pip install simsimd`."
        )
        X_norm = np.linalg.norm(X, axis=1)
        Y_norm = np.linalg.norm(Y, axis=1)
        # Ignore divide by zero errors run time warnings as those are handled below.
        with np.errstate(divide="ignore", invalid="ignore"):
            similarity = np.dot(X, Y.T) / np.outer(X_norm, Y_norm)
        similarity[np.isnan(similarity) | np.isinf(similarity)] = 0.0
        return similarity


def maximal_marginal_relevance(
    query_embedding: np.ndarray,
    embedding_list: list,
    lambda_mult: float = 0.5,
    k: int = 4,
) -> List[int]:
    r"""Compute Maximal Marginal Relevance (MMR).

    MMR is a technique used to select documents that are both relevant to the query
    and diverse among themselves. This function returns the indices
    of the top-k embeddings that maximize the marginal relevance.

    Args:
        query_embedding (np.ndarray): The embedding vector of the query.
        embedding_list (list of np.ndarray): A list containing the embedding vectors
            of the candidate documents.
        lambda_mult (float, optional): The trade-off parameter between
            relevance and diversity. Defaults to 0.5.
        k (int, optional): The number of embeddings to select. Defaults to 4.

    Returns:
        list of int: The indices of the embeddings that maximize the marginal relevance.

    Notes:
        The Maximal Marginal Relevance (MMR) is computed using the following formula:

    MMR = argmax_{D_i ∈ R \ S} [λ * Sim(D_i, Q) - (1 - λ) * max_{D_j ∈ S} Sim(D_i, D_j)]

        where:
        - R is the set of candidate documents,
        - S is the set of selected documents,
        - Q is the query embedding,
        - Sim(D_i, Q) is the similarity between document D_i and the query,
        - Sim(D_i, D_j) is the similarity between documents D_i and D_j,
        - λ is the trade-off parameter.
    """

    if min(k, len(embedding_list)) <= 0:
        return []
    if query_embedding.ndim == 1:
        query_embedding = np.expand_dims(query_embedding, axis=0)
    similarity_to_query = cosine_similarity(query_embedding, embedding_list)[0]
    most_similar = int(np.argmax(similarity_to_query))
    idxs = [most_similar]
    selected = np.array([embedding_list[most_similar]])
    while len(idxs) < min(k, len(embedding_list)):
        best_score = -np.inf
        idx_to_add = -1
        similarity_to_selected = cosine_similarity(embedding_list, selected)
        for i, query_score in enumerate(similarity_to_query):
            if i in idxs:
                continue
            redundant_score = max(similarity_to_selected[i])
            equation_score = (
                lambda_mult * query_score - (1 - lambda_mult) * redundant_score
            )
            if equation_score > best_score:
                best_score = equation_score
                idx_to_add = i
        idxs.append(idx_to_add)
        selected = np.append(selected, [embedding_list[idx_to_add]], axis=0)
    return idxs


def str_to_oid(str_repr: str) -> Any | str:
    """Attempt to cast string representation of id to MongoDB's internal BSON ObjectId.

    To be consistent with ObjectId, input must be a 24 character hex string.
    If it is not, MongoDB will happily use the string in the main _id index.
    Importantly, the str representation that comes out of MongoDB will have this form.

    Args:
        str_repr: id as string.

    Returns:
        ObjectID
    """
    from bson import ObjectId
    from bson.errors import InvalidId

    try:
        return ObjectId(str_repr)
    except InvalidId:
        logger.debug(
            "ObjectIds must be 12-character byte or 24-character hex strings. "
            "Examples: b'heres12bytes', '6f6e6568656c6c6f68656768'"
        )
        return str_repr


def oid_to_str(oid: Any) -> str:
    """Convert MongoDB's internal BSON ObjectId into a simple str for compatibility.

    Instructive helper to show where data is coming out of MongoDB.

    Args:
        oid: bson.ObjectId

    Returns:
        24 character hex string.
    """
    return str(oid)


def make_serializable(
    obj: Dict[str, Any],
) -> None:
    """Recursively cast values in a dict to a form able to json.dump"""

    from bson import ObjectId

    for k, v in obj.items():
        if isinstance(v, dict):
            make_serializable(v)
        elif isinstance(v, list) and v and isinstance(v[0], (ObjectId, date, datetime)):
            obj[k] = [oid_to_str(item) for item in v]
        elif isinstance(v, ObjectId):
            obj[k] = oid_to_str(v)
        elif isinstance(v, (datetime, date)):
            obj[k] = v.isoformat()



from typing import Type

from langchain_tests.integration_tests import ToolsIntegrationTests

from langchain_mongodb.agent_toolkit.tool import (
    InfoMongoDBDatabaseTool,
    ListMongoDBDatabaseTool,
    MongoDBDatabase,
    QueryMongoDBCheckerTool,
    QueryMongoDBDatabaseTool,
)
from tests.utils import create_database, create_llm


class TestQueryMongoDBDatabaseToolIntegration(ToolsIntegrationTests):
    _db: MongoDBDatabase

    @classmethod
    def setup_class(cls):
        cls._db = create_database()

    @classmethod
    def teardown_class(cls):
        cls._db.close()

    @property
    def tool_constructor(self) -> Type[QueryMongoDBDatabaseTool]:
        return QueryMongoDBDatabaseTool

    @property
    def tool_constructor_params(self) -> dict:
        return dict(db=self._db)

    @property
    def tool_invoke_params_example(self) -> dict:
        return dict(query='db.test.aggregate([{"$match": {}}])')


class TestInfoMongoDBDatabaseToolIntegration(ToolsIntegrationTests):
    _db: MongoDBDatabase

    @classmethod
    def setup_class(cls):
        cls._db = create_database()

    @classmethod
    def teardown_class(cls):
        cls._db.close()

    @property
    def tool_constructor(self) -> Type[InfoMongoDBDatabaseTool]:
        return InfoMongoDBDatabaseTool

    @property
    def tool_constructor_params(self) -> dict:
        return dict(db=self._db)

    @property
    def tool_invoke_params_example(self) -> dict:
        return dict(collection_names="test")


class TestListMongoDBDatabaseToolIntegration(ToolsIntegrationTests):
    _db: MongoDBDatabase

    @classmethod
    def setup_class(cls):
        cls._db = create_database()

    @classmethod
    def teardown_class(cls):
        cls._db.close()

    @property
    def tool_constructor(self) -> Type[ListMongoDBDatabaseTool]:
        return ListMongoDBDatabaseTool

    @property
    def tool_constructor_params(self) -> dict:
        return dict(db=self._db)

    @property
    def tool_invoke_params_example(self) -> dict:
        return dict()


class TestQueryMongoDBCheckerToolIntegration(ToolsIntegrationTests):
    _db: MongoDBDatabase

    @classmethod
    def setup_class(cls):
        cls._db = create_database()

    @classmethod
    def teardown_class(cls):
        cls._db.close()

    @property
    def tool_constructor(self) -> Type[QueryMongoDBCheckerTool]:
        return QueryMongoDBCheckerTool

    @property
    def tool_constructor_params(self) -> dict:
        return dict(db=self._db, llm=create_llm())

    @property
    def tool_invoke_params_example(self) -> dict:
        return dict(query='db.test.aggregate([{"$match": {}}])')



"""Test MongoDBAtlasVectorSearch.from_documents."""

from __future__ import annotations

import pytest  # type: ignore[import-not-found]
from langchain_core.vectorstores import VectorStore
from langchain_tests.integration_tests import VectorStoreIntegrationTests
from pymongo import MongoClient
from pymongo.collection import Collection

from langchain_mongodb.index import (
    create_vector_search_index,
)

from ..utils import DB_NAME, PatchedMongoDBAtlasVectorSearch

COLLECTION_NAME = "langchain_test_standard"
INDEX_NAME = "langchain-test-index-standard"
DIMENSIONS = 6


@pytest.fixture
def collection(client: MongoClient) -> Collection:
    if COLLECTION_NAME not in client[DB_NAME].list_collection_names():
        clxn = client[DB_NAME].create_collection(COLLECTION_NAME)
    else:
        clxn = client[DB_NAME][COLLECTION_NAME]

    clxn.delete_many({})

    if not any([INDEX_NAME == ix["name"] for ix in clxn.list_search_indexes()]):
        create_vector_search_index(
            collection=clxn,
            index_name=INDEX_NAME,
            dimensions=DIMENSIONS,
            path="embedding",
            filters=["c"],
            similarity="cosine",
            wait_until_complete=60,
        )

    return clxn


class TestMongoDBAtlasVectorSearch(VectorStoreIntegrationTests):
    @pytest.fixture()
    def vectorstore(self, collection) -> VectorStore:  # type: ignore
        """Get an empty vectorstore for unit tests."""
        store = PatchedMongoDBAtlasVectorSearch(
            collection, self.get_embeddings(), index_name=INDEX_NAME
        )
        # note: store should be EMPTY at this point
        # if you need to delete data, you may do so here
        return store



import os
import sqlite3

import pytest
import requests
from flaky import flaky  # type:ignore[import-untyped]
from langchain.agents import create_agent  # type: ignore[assignment]
from langchain_openai import AzureChatOpenAI, ChatOpenAI
from pymongo import MongoClient

from langchain_mongodb.agent_toolkit import (
    MONGODB_AGENT_SYSTEM_PROMPT,
    MongoDBDatabase,
    MongoDBDatabaseToolkit,
)
from tests.utils import CONNECTION_STRING

DB_NAME = "langchain_test_db_chinook"


@pytest.fixture
def db(client: MongoClient) -> MongoDBDatabase:
    # Load the raw data into sqlite.
    url = "https://raw.githubusercontent.com/lerocha/chinook-database/master/ChinookDatabase/DataSources/Chinook_Sqlite.sql"
    response = requests.get(url)
    sql_script = response.text
    con = sqlite3.connect(":memory:", check_same_thread=False)
    con.executescript(sql_script)

    # Convert the sqlite data to MongoDB data.
    con.row_factory = sqlite3.Row
    cursor = con.cursor()
    sql_query = """SELECT name FROM sqlite_master WHERE type='table';"""
    cursor.execute(sql_query)
    tables = [i[0] for i in cursor.fetchall()]
    cursor.close()
    for t in tables:
        coll = client[DB_NAME][t]
        coll.delete_many({})
        cursor = con.cursor()
        cursor.execute(f"select * from {t}")
        docs = [dict(i) for i in cursor.fetchall()]
        cursor.close()
        coll.insert_many(docs)
    return MongoDBDatabase(client, DB_NAME)


@flaky(max_runs=5, min_passes=4)
@pytest.mark.skipif(
    "OPENAI_API_KEY" not in os.environ and "AZURE_OPENAI_ENDPOINT" not in os.environ,
    reason="test requires OpenAI for chat responses",
)
def test_toolkit_response(db):
    db_wrapper = MongoDBDatabase.from_connection_string(
        CONNECTION_STRING, database=DB_NAME
    )
    if "AZURE_OPENAI_ENDPOINT" in os.environ:
        llm = AzureChatOpenAI(model="gpt-4o-mini", timeout=60, seed=12345)
    else:
        llm = ChatOpenAI(model="gpt-4o-mini", timeout=60, seed=12345)

    toolkit = MongoDBDatabaseToolkit(db=db_wrapper, llm=llm)

    prompt = MONGODB_AGENT_SYSTEM_PROMPT.format(top_k=5)

    test_query = "Which country's customers spent the most?"
    agent = create_agent(llm, toolkit.get_tools(), system_prompt=prompt)
    agent.step_timeout = 60
    events = agent.stream(
        {"messages": [("user", test_query)]},
        stream_mode="values",
    )
    messages = []
    for event in events:
        messages.extend(event["messages"])
    assert "USA" in messages[-1].content, messages[-1].content
    db_wrapper.close()



import json
import warnings

import pytest
from langchain_classic.memory import (
    ConversationBufferMemory,
)
from langchain_core.messages import message_to_dict
from pytest_mock import MockerFixture

from langchain_mongodb.chat_message_histories import MongoDBChatMessageHistory
from langchain_mongodb.utils import DRIVER_METADATA

from ..utils import MockCollection


class PatchedMongoDBChatMessageHistory(MongoDBChatMessageHistory):
    def __init__(self) -> None:
        self.session_id = "test-session"
        self.database_name = "test-database"
        self.collection_name = "test-collection"
        self.collection = MockCollection()
        self.session_id_key = "SessionId"
        self.history_key = "History"
        self.history_size = None
        self.db = self.collection.database
        self.client = self.db.client


def test_memory_with_message_store() -> None:
    """Test the memory with a message store."""
    # setup MongoDB as a message store
    message_history = PatchedMongoDBChatMessageHistory()
    with warnings.catch_warnings():
        # Ignore warnings raised by base class.
        warnings.simplefilter("ignore", DeprecationWarning)
        memory = ConversationBufferMemory(
            memory_key="baz", chat_memory=message_history, return_messages=True
        )

    # add some messages
    memory.chat_memory.add_ai_message("This is me, the AI")
    memory.chat_memory.add_user_message("This is me, the human")

    # get the message history from the memory store and turn it into a json
    messages = memory.chat_memory.messages
    messages_json = json.dumps([message_to_dict(msg) for msg in messages])

    assert "This is me, the AI" in messages_json
    assert "This is me, the human" in messages_json

    # remove the record from MongoDB, so the next test run won't pick it up
    memory.chat_memory.clear()

    assert memory.chat_memory.messages == []

    message_history.close()
    assert message_history.client.is_closed


def test_init_with_connection_string(mocker: MockerFixture) -> None:
    mock_mongo_client = mocker.patch(
        "langchain_mongodb.chat_message_histories.MongoClient"
    )

    history = MongoDBChatMessageHistory(
        connection_string="mongodb://localhost:27017/",
        session_id="test-session",
        database_name="test-database",
        collection_name="test-collection",
    )

    mock_mongo_client.assert_called_once_with(
        "mongodb://localhost:27017/",
        driver=DRIVER_METADATA,
    )
    assert history.session_id == "test-session"
    assert history.database_name == "test-database"
    assert history.collection_name == "test-collection"
    history.close()


def test_init_raises_error_without_connection_or_client() -> None:
    with pytest.raises(
        ValueError, match="Either connection_string or client must be provided"
    ):
        MongoDBChatMessageHistory(
            session_id="test_session",
            connection_string=None,
            client=None,
        )



from typing import Generator, List, Optional

import pytest
from pymongo import MongoClient
from pymongo.collection import Collection

from langchain_mongodb import MongoDBAtlasVectorSearch, index

from ..utils import ConsistentFakeEmbeddings

DB_NAME = "langchain_test_index_db"
COLLECTION_NAME = "test_index"
VECTOR_INDEX_NAME = "vector_index"

TIMEOUT = 120
DIMENSIONS = 10


@pytest.fixture
def collection(client: MongoClient) -> Generator:
    """Depending on uri, this could point to any type of cluster."""
    if COLLECTION_NAME not in client[DB_NAME].list_collection_names():
        clxn = client[DB_NAME].create_collection(COLLECTION_NAME)
    else:
        clxn = client[DB_NAME][COLLECTION_NAME]
    clxn = client[DB_NAME][COLLECTION_NAME]
    clxn.delete_many({})
    yield clxn
    clxn.delete_many({})


def test_search_index_drop_add_delete_commands(collection: Collection) -> None:
    index_name = VECTOR_INDEX_NAME
    dimensions = DIMENSIONS
    path = "embedding"
    similarity = "cosine"
    filters: Optional[List[str]] = None
    wait_until_complete = TIMEOUT

    for index_info in collection.list_search_indexes():
        index.drop_vector_search_index(
            collection, index_info["name"], wait_until_complete=wait_until_complete
        )

    assert len(list(collection.list_search_indexes())) == 0

    index.create_vector_search_index(
        collection=collection,
        index_name=index_name,
        dimensions=dimensions,
        path=path,
        similarity=similarity,
        filters=filters,
        wait_until_complete=wait_until_complete,
    )

    assert index._is_index_ready(collection, index_name)
    indexes = list(collection.list_search_indexes())
    assert len(indexes) == 1
    assert indexes[0]["name"] == index_name

    index.drop_vector_search_index(
        collection, index_name, wait_until_complete=wait_until_complete
    )

    indexes = list(collection.list_search_indexes())
    assert len(indexes) == 0


@pytest.mark.skip("collection.update_vector_search_index requires [CLOUDP-275518]")
def test_search_index_update_vector_search_index(collection: Collection) -> None:
    index_name = "INDEX_TO_UPDATE"
    similarity_orig = "cosine"
    similarity_new = "euclidean"

    # Create another index
    index.create_vector_search_index(
        collection=collection,
        index_name=index_name,
        dimensions=DIMENSIONS,
        path="embedding",
        similarity=similarity_orig,
        wait_until_complete=TIMEOUT,
    )

    assert index._is_index_ready(collection, index_name)
    indexes = list(collection.list_search_indexes())
    assert len(indexes) == 1
    assert indexes[0]["name"] == index_name
    assert indexes[0]["latestDefinition"]["fields"][0]["similarity"] == similarity_orig

    # Update the index and test new similarity
    index.update_vector_search_index(
        collection=collection,
        index_name=index_name,
        dimensions=DIMENSIONS,
        path="embedding",
        similarity=similarity_new,
        wait_until_complete=TIMEOUT,
    )

    assert index._is_index_ready(collection, index_name)
    indexes = list(collection.list_search_indexes())
    assert len(indexes) == 1
    assert indexes[0]["name"] == index_name
    assert indexes[0]["latestDefinition"]["fields"][0]["similarity"] == similarity_new


def test_vectorstore_create_vector_search_index(collection: Collection) -> None:
    """Tests vectorstore wrapper around index command."""

    # Set up using the index module's api
    if len(list(collection.list_search_indexes())) != 0:
        index.drop_vector_search_index(
            collection, VECTOR_INDEX_NAME, wait_until_complete=TIMEOUT
        )

    # Test MongoDBAtlasVectorSearch's API
    _ = MongoDBAtlasVectorSearch(
        collection=collection,
        embedding=ConsistentFakeEmbeddings(),
        index_name=VECTOR_INDEX_NAME,
        dimensions=DIMENSIONS,
        auto_index_timeout=TIMEOUT,
    )

    assert index._is_index_ready(collection, VECTOR_INDEX_NAME)
    indexes = list(collection.list_search_indexes())
    assert len(indexes) == 1
    assert indexes[0]["name"] == VECTOR_INDEX_NAME

    # Tear down using the index module's api
    index.drop_vector_search_index(
        collection, VECTOR_INDEX_NAME, wait_until_complete=TIMEOUT
    )



from typing import List

from langchain_core.documents import Document
from pymongo import MongoClient

from langchain_mongodb.docstores import MongoDBDocStore

from ..utils import DB_NAME

COLLECTION_NAME = "langchain_test_docstore"


def test_docstore(client: MongoClient, technical_report_pages: List[Document]) -> None:
    db = client[DB_NAME]
    db.drop_collection(COLLECTION_NAME)
    clxn = db[COLLECTION_NAME]

    n_docs = len(technical_report_pages)
    assert clxn.count_documents({}) == 0
    docstore = MongoDBDocStore(collection=clxn)

    docstore.mset([(str(i), technical_report_pages[i]) for i in range(n_docs)])
    assert clxn.count_documents({}) == n_docs

    twenties = list(docstore.yield_keys(prefix="2"))
    assert len(twenties) == 11  # includes 2, 20, 21, ..., 29

    docstore.mdelete([str(i) for i in range(20, 30)] + ["2"])
    assert clxn.count_documents({}) == n_docs - 11
    assert set(docstore.mget(twenties)) == {None}

    sample = docstore.mget(["8", "16", "24", "36"])
    assert sample[2] is None
    assert all(isinstance(sample[i], Document) for i in [0, 1, 3])



"""Test MongoDBAtlasVectorSearch.from_documents."""

from __future__ import annotations

from typing import List

import pytest  # type: ignore[import-not-found]
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from pymongo import MongoClient
from pymongo.collection import Collection

from langchain_mongodb.index import (
    create_vector_search_index,
)

from ..utils import DB_NAME, ConsistentFakeEmbeddings, PatchedMongoDBAtlasVectorSearch

COLLECTION_NAME = "langchain_test_from_documents"
INDEX_NAME = "langchain-test-index-from-documents"
DIMENSIONS = 5


@pytest.fixture(scope="module")
def collection(client: MongoClient) -> Collection:
    if COLLECTION_NAME not in client[DB_NAME].list_collection_names():
        clxn = client[DB_NAME].create_collection(COLLECTION_NAME)
    else:
        clxn = client[DB_NAME][COLLECTION_NAME]

    clxn.delete_many({})

    if not any([INDEX_NAME == ix["name"] for ix in clxn.list_search_indexes()]):
        create_vector_search_index(
            collection=clxn,
            index_name=INDEX_NAME,
            dimensions=DIMENSIONS,
            path="embedding",
            similarity="cosine",
            wait_until_complete=60,
        )

    return clxn


@pytest.fixture(scope="module")
def example_documents() -> List[Document]:
    return [
        Document(page_content="Dogs are tough.", metadata={"a": 1}),
        Document(page_content="Cats have fluff.", metadata={"b": 1}),
        Document(page_content="What is a sandwich?", metadata={"c": 1}),
        Document(page_content="That fence is purple.", metadata={"d": 1, "e": 2}),
    ]


@pytest.fixture(scope="module")
def embeddings() -> Embeddings:
    return ConsistentFakeEmbeddings(DIMENSIONS)


@pytest.fixture(scope="module")
def vectorstore(
    collection: Collection, example_documents: List[Document], embeddings: Embeddings
) -> PatchedMongoDBAtlasVectorSearch:
    """VectorStore created with a few documents and a trivial embedding model.

    Note: PatchedMongoDBAtlasVectorSearch is MongoDBAtlasVectorSearch in all
    but one important feature. It waits until all documents are fully indexed
    before returning control to the caller.
    """
    return PatchedMongoDBAtlasVectorSearch.from_documents(
        example_documents,
        embedding=embeddings,
        collection=collection,
        index_name=INDEX_NAME,
    )


def test_default_search(
    vectorstore: PatchedMongoDBAtlasVectorSearch, example_documents: List[Document]
) -> None:
    """Test end to end construction and search."""
    output = vectorstore.similarity_search("Sandwich", k=1)
    assert len(output) == 1
    # Check for the presence of the metadata key
    assert any(
        [key.page_content == output[0].page_content for key in example_documents]
    )
    # Assert no presence of embeddings in results
    assert all(["embedding" not in key.metadata for key in output])


def test_search_with_embeddings(vectorstore: PatchedMongoDBAtlasVectorSearch) -> None:
    output = vectorstore.similarity_search("Sandwich", k=2, include_embeddings=True)
    assert len(output) == 2

    # Assert embeddings in results
    assert all([key.metadata.get("embedding") for key in output])



import os
import uuid
from typing import Any, List, Union

import pytest  # type: ignore[import-not-found]
from langchain_core.caches import BaseCache
from langchain_core.globals import (
    get_llm_cache,
    set_llm_cache,
)
from langchain_core.load.dump import dumps
from langchain_core.messages import AIMessage, BaseMessage, HumanMessage
from langchain_core.outputs import ChatGeneration, Generation, LLMResult
from pymongo import MongoClient
from pymongo.collection import Collection

from langchain_mongodb.cache import MongoDBAtlasSemanticCache, MongoDBCache
from langchain_mongodb.index import (
    create_vector_search_index,
)

from ..utils import DB_NAME, ConsistentFakeEmbeddings, FakeChatModel, FakeLLM

CONN_STRING = os.environ.get("MONGODB_URI")
INDEX_NAME = "langchain-test-index-semantic-cache"
COLLECTION = "langchain_test_cache"

DIMENSIONS = 1536  # Meets OpenAI model
TIMEOUT = 60.0


def random_string() -> str:
    return str(uuid.uuid4())


@pytest.fixture(scope="module")
def collection(client: MongoClient) -> Collection:
    """A Collection with both a Vector and a Full-text Search Index"""
    if COLLECTION not in client[DB_NAME].list_collection_names():
        clxn = client[DB_NAME].create_collection(COLLECTION)
    else:
        clxn = client[DB_NAME][COLLECTION]

    clxn.delete_many({})

    if not any([INDEX_NAME == ix["name"] for ix in clxn.list_search_indexes()]):
        create_vector_search_index(
            collection=clxn,
            index_name=INDEX_NAME,
            dimensions=DIMENSIONS,
            path="embedding",
            filters=["llm_string"],
            similarity="cosine",
            wait_until_complete=TIMEOUT,
        )

    return clxn


def llm_cache(cls: Any) -> BaseCache:
    set_llm_cache(
        cls(
            embedding=ConsistentFakeEmbeddings(dimensionality=DIMENSIONS),
            connection_string=CONN_STRING,
            collection_name=COLLECTION,
            database_name=DB_NAME,
            index_name=INDEX_NAME,
            score_threshold=0.5,
            wait_until_ready=TIMEOUT,
        )
    )
    assert get_llm_cache()
    return get_llm_cache()  # type:ignore[return-value]


@pytest.fixture(scope="module", autouse=True)
def reset_cache():
    """Prevents global cache being affected in other module's tests."""
    yield
    print("\nAll cache tests have finished. Setting global cache to None.")
    set_llm_cache(None)


def _execute_test(
    prompt: Union[str, List[BaseMessage]],
    llm: Union[str, FakeLLM, FakeChatModel],
    response: List[Generation],
) -> None:
    # Fabricate an LLM String

    if not isinstance(llm, str):
        params = llm.dict()
        params["stop"] = None
        llm_string = str(sorted([(k, v) for k, v in params.items()]))
    else:
        llm_string = llm

    # If the prompt is a str then we should pass just the string
    dumped_prompt: str = prompt if isinstance(prompt, str) else dumps(prompt)

    # Update the cache
    get_llm_cache().update(dumped_prompt, llm_string, response)  # type:ignore[union-attr]

    # Retrieve the cached result through 'generate' call
    output: Union[List[Generation], LLMResult, None]
    expected_output: Union[List[Generation], LLMResult]

    if isinstance(llm, str):
        output = get_llm_cache().lookup(dumped_prompt, llm)  # type: ignore
        expected_output = response
    else:
        output = llm.generate([prompt])  # type: ignore
        expected_output = LLMResult(
            generations=[response],
            llm_output={},
        )

    # Work around older output with "usage_metadata"
    if "usage_metadata" in str(output) and "usage_metadata" not in str(expected_output):
        for generation in output.generations[0]:  # type: ignore
            generation.message.usage_metadata = None  # type: ignore

    assert output == expected_output  # type: ignore


@pytest.mark.parametrize(
    "prompt, llm, response",
    [
        ("foo", "bar", [Generation(text="fizz")]),
        ("foo", FakeLLM(), [Generation(text="fizz")]),
        (
            [HumanMessage(content="foo")],
            FakeChatModel(),
            [ChatGeneration(message=AIMessage(content="foo"))],
        ),
    ],
    ids=[
        "plain_cache",
        "cache_with_llm",
        "cache_with_chat",
    ],
)
@pytest.mark.parametrize("cacher", [MongoDBCache, MongoDBAtlasSemanticCache])
@pytest.mark.parametrize("remove_score", [True, False])
def test_mongodb_cache(
    remove_score: bool,
    cacher: Union[MongoDBCache, MongoDBAtlasSemanticCache],
    prompt: Union[str, List[BaseMessage]],
    llm: Union[str, FakeLLM, FakeChatModel],
    response: List[Generation],
    collection: Collection,
) -> None:
    llm_cache(cacher)
    if remove_score:
        get_llm_cache().score_threshold = None  # type: ignore
    try:
        _execute_test(prompt, llm, response)
    finally:
        get_llm_cache().clear()  # type:ignore[union-attr]
        get_llm_cache().close()  # type:ignore[attr-defined,union-attr]


@pytest.mark.parametrize(
    "prompts, generations",
    [
        # Single prompt, single generation
        ([random_string()], [[random_string()]]),
        # Single prompt, multiple generations
        ([random_string()], [[random_string(), random_string()]]),
        # Single prompt, multiple generations
        ([random_string()], [[random_string(), random_string(), random_string()]]),
        # Multiple prompts, multiple generations
        (
            [random_string(), random_string()],
            [[random_string()], [random_string(), random_string()]],
        ),
    ],
    ids=[
        "single_prompt_single_generation",
        "single_prompt_two_generations",
        "single_prompt_three_generations",
        "multiple_prompts_multiple_generations",
    ],
)
def test_mongodb_atlas_cache_matrix(
    prompts: List[str],
    generations: List[List[str]],
    collection: Collection,
) -> None:
    llm_cache(MongoDBAtlasSemanticCache)
    llm = FakeLLM()

    # Fabricate an LLM String
    params = llm.dict()
    params["stop"] = None
    llm_string = str(sorted([(k, v) for k, v in params.items()]))

    llm_generations = [
        [
            Generation(text=generation, generation_info=params)
            for generation in prompt_i_generations
        ]
        for prompt_i_generations in generations
    ]

    for prompt_i, llm_generations_i in zip(prompts, llm_generations, strict=True):
        _execute_test(prompt_i, llm_string, llm_generations_i)
    assert llm.generate(prompts) == LLMResult(
        generations=llm_generations, llm_output={}
    )
    get_llm_cache().clear()  # type:ignore[union-attr]
    get_llm_cache().close()  # type:ignore[attr-defined,union-attr]



from time import sleep
from typing import Generator

import pytest
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.errors import OperationFailure, ServerSelectionTimeoutError

from langchain_mongodb import index

DIMENSION = 5
TIMEOUT = 120


@pytest.fixture
def collection() -> Generator[Collection, None, None]:
    """Collection on MongoDB Cluster, not an Atlas one."""
    client: MongoClient = MongoClient()
    yield client["db"]["collection"]
    client.close()


def test_create_vector_search_index(collection: Collection) -> None:
    with pytest.raises((OperationFailure, ServerSelectionTimeoutError)):
        index.create_vector_search_index(
            collection,
            "index_name",
            DIMENSION,
            "embedding",
            "cosine",
            [],
            wait_until_complete=TIMEOUT,
        )


def test_drop_vector_search_index(collection: Collection) -> None:
    with pytest.raises((OperationFailure, ServerSelectionTimeoutError)):
        index.drop_vector_search_index(
            collection, "index_name", wait_until_complete=TIMEOUT
        )


def test_update_vector_search_index(collection: Collection) -> None:
    with pytest.raises((OperationFailure, ServerSelectionTimeoutError)):
        index.update_vector_search_index(
            collection,
            "index_name",
            DIMENSION,
            "embedding",
            "cosine",
            [],
            wait_until_complete=TIMEOUT,
        )


def test___is_index_ready(collection: Collection) -> None:
    with pytest.raises((OperationFailure, ServerSelectionTimeoutError)):
        index._is_index_ready(collection, "index_name")


def test__wait_for_predicate() -> None:
    err = "error string"
    with pytest.raises(TimeoutError) as e:
        index._wait_for_predicate(lambda: sleep(5), err=err, timeout=0.5, interval=0.1)
        assert err in str(e)

    index._wait_for_predicate(lambda: True, err=err, timeout=1.0, interval=0.5)



import pytest
from langchain_text_splitters import RecursiveCharacterTextSplitter

from langchain_mongodb.docstores import MongoDBDocStore
from langchain_mongodb.retrievers import (
    MongoDBAtlasFullTextSearchRetriever,
    MongoDBAtlasHybridSearchRetriever,
    MongoDBAtlasParentDocumentRetriever,
)
from langchain_mongodb.vectorstores import MongoDBAtlasVectorSearch

from ..utils import ConsistentFakeEmbeddings, MockCollection


@pytest.fixture()
def collection() -> MockCollection:
    return MockCollection()


@pytest.fixture()
def embeddings() -> ConsistentFakeEmbeddings:
    return ConsistentFakeEmbeddings()


def test_full_text_search(collection):
    search = MongoDBAtlasFullTextSearchRetriever(
        collection=collection, search_index_name="foo", search_field="bar"
    )
    search.close()
    assert collection.database.client.is_closed


def test_hybrid_search(collection, embeddings):
    vs = MongoDBAtlasVectorSearch(collection, embeddings)
    search = MongoDBAtlasHybridSearchRetriever(vectorstore=vs, search_index_name="foo")
    search.close()
    assert collection.database.client.is_closed


def test_parent_retriever(collection, embeddings):
    vs = MongoDBAtlasVectorSearch(collection, embeddings)
    ds = MongoDBDocStore(collection)
    cs = RecursiveCharacterTextSplitter(chunk_size=400)
    retriever = MongoDBAtlasParentDocumentRetriever(
        vectorstore=vs, docstore=ds, child_splitter=cs
    )
    retriever.close()
    assert collection.database.client.is_closed



from langchain_core.prompts.chat import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)

ENTITY_EXTRACTION_INSTRUCTIONS = """
## Overview
You are a meticulous analyst tasked with identifying potential entities from unstructured text
to build a knowledge graph in a structured json format of entities (nodes) and their relationships (edges).
**Include as many entities and relationships as you can.**

INPUT: You will be provided a text document.
OUTPUT:
- You will produce valid json according the "Output Schema" section below.
- Your response **must be** a **valid JSON document** with NO extra text, explanations, or markdown formatting.
- The extracted entities and relationships **MUST STRICTLY CONFORM** to the constraints outlined below.
- Any entities or relationships not matching the allowed types must be **EXCLUDED**.


## Entities
An entity in a knowledge graph is a uniquely identifiable object or concept
(such as a person, organization, location, object, or event),
represented as a node with attributes (properties) and relationships to other entities.

Use the reserved field name `_id` for the name. It will be a unique primary key,
and MongoDB automatically creates an index for the `_id` field.

Maintain Entity Consistency when extracting entities. If an entity, such as "John Doe",
is mentioned multiple times in the text but is referred to by different names or pronouns (e.g., "John", "Mr Doe", "he"),
always use the most complete identifier for that entity throughout the knowledge graph.
In this example, use "John Doe" as the entity `_id.`

**Allowed Entity Types**:
- Extract ONLY entities whose `type` matches one of the following: {allowed_entity_types}.
- NOTE: If this list is empty, ANY `type` is permitted.

### Examples of Exclusions:
- If `allowed_entity_types` is `["Person", "Organization"]`, and the text mentions "Event" or "Location",
  these entities must **NOT** be included in the output.

## Relationships
Relationships represent edges in the knowledge graph. Relationships describe a specific edge type.
Relationships MUST include a target entity, but Entities can be extracted that DO NOT have relationships!
Ensure consistency and generality in relationship names when constructing knowledge schemas.
Instead of using specific and momentary types such as 'worked_at', use more general and timeless relationship types
like 'employee'. Add details as attributes. Make sure to use general and timeless relationship types!

### CRITICAL: Array Length Alignment
The relationships object contains three arrays: `target_ids`, `types`, and `attributes`.
**These three arrays MUST have EXACTLY the same length.**
- Each position (index) in these arrays describes ONE complete relationship.
- Position 0 in `target_ids`, `types`, and `attributes` together describe the first relationship.
- Position 1 in `target_ids`, `types`, and `attributes` together describe the second relationship.
- And so on...

If a relationship has no attributes, you MUST still include an empty object `{{}}` in the `attributes` array at that position.

Example of CORRECT alignment:
```json
"relationships": {{
  "target_ids": ["Entity A", "Entity B"],
  "types": ["partners", "supplier"],
  "attributes": [
    {{"since": ["2020"]}},
    {{}}
  ]
}}
```

Example of INCORRECT (DO NOT DO THIS):
```json
"relationships": {{
  "target_ids": ["Entity A", "Entity B"],
  "types": ["partners"],
  "attributes": [{{"since": ["2020"]}}]
}}
```

**Allowed Relationship Types**:
- Extract ONLY relationships whose `type` matches one of the following: {allowed_relationship_types}.
- If this list is empty, ANY relationship type is permitted.
- Map synonymous or related terms to the closest matching allowed type. For example:
	-	“works for” or “employed by” → employee
	-	“manages” or “supervises” → manager
- If a relationship cannot be named with one of the allowed keys, **DO NOT include it**.
- An entity need not have a relationships object if no relationship is found that matches the allowed relation types.

### Examples of Exclusions:
- If `allowed_relationship_types` is `["employs", "friend"]` and the text implies a "partner" relationship,
  the entities can be added, but the "partner" relationship must **NOT** be included.

## Validation
Before producing the final output:
1. Validate that all extracted entities have an `_id` and `type`.
2. Validate that all `type` values are in {allowed_entity_types}.
3. Validate that all relationships use keys in {allowed_relationship_types}.
4. **CRITICAL**: For each entity with relationships, verify that `target_ids`, `types`, and `attributes` arrays have EXACTLY the same length.
5. Exclude any entities or relationships failing validation.

## Output Schema
Output a valid JSON document with a single top-level key, `entities`, as an array of objects.
Each object must conform to the following schema:
{entity_schema}

{entity_examples}
"""


NAME_EXTRACTION_INSTRUCTIONS = """
You are an analyst tasked with identifying potential entities in text documents.
You will be provided a short document from which you infer entity names.
Identify as many as possible.

Provide your response as a valid JSON Array of entity names
or human-readable identifiers, found in the text.

**Allowed Entity Types**:
- By default, all types are permitted.
- If a non-empty list is provided, extract ONLY entities whose `type` matches one of the following: [{allowed_entity_types}].

### Examples of Exclusions:
- If `allowed_entity_types` is `["Person", "Organization"]`, and the text mentions an "Event" or "Location",
  these entities must **NOT** be included in the output.

 ## Examples:
 Example 1: `allowed_entity_types` is `[]`
    input: "John Doe works at ACME in New York"
    output: ["John Doe", "ACME", "New York"]

 In this example, you would identify 3 entities:
 John Doe of type person; ACME of type organization; New York of type place.

Example 2: `allowed_entity_types` is `[organization, place]`
    input: "John Doe works at ACME in New York"
    output: ["ACME", "New York"]

 In this example, you would identify only 2 entities:
 ACME of type organization; New York of type place.
 John Doe, of type person, would be excluded.

 2. input: "In what continent is Brazil?
    output: ["Brazil"]

This example is in the form of a question. There is one entity,

3. input: "For legal and operational purposes, many governments and organizations adopt specific definitions."
   output: []

In the third example, there are no entities.
Though there are concepts and nouns that might be types or attributes of entities,
there is nothing here that could be seen as being a unique identifier or name.

4. input: ""
   output: []

In final third example, there are no entities.

### (Optional) Additional Examples

{entity_name_examples}
"""


RAG_INSTRUCTIONS = """
## Context
You are a meticulous analyst tasked with extracting information in the form of knowledge graphs
comprised of entities (nodes) and their relationships (edges).

Based on the user input (query) that will be provided, you have already retrieved information
from the knowledge graph in the form of a list of entities known to be related to those in the Query.

From the context retrieved alone, please respond to the Query.
Your response should be a string of concise prose.

## Entity Schema
The entities have the following schema matching MongoDB's $jsonSchema style used for MongoDB validation.

{entity_schema}

## Entities Found to be Related to Query
{related_entities}
"""

entity_prompt = ChatPromptTemplate.from_messages(
    [
        SystemMessagePromptTemplate.from_template(ENTITY_EXTRACTION_INSTRUCTIONS),
        HumanMessagePromptTemplate.from_template("{input_document}"),
    ]
)

query_prompt = ChatPromptTemplate.from_messages(
    [
        SystemMessagePromptTemplate.from_template(NAME_EXTRACTION_INSTRUCTIONS),
        HumanMessagePromptTemplate.from_template("{input_document}"),
    ]
)

rag_prompt = ChatPromptTemplate.from_messages(
    [
        SystemMessagePromptTemplate.from_template(RAG_INSTRUCTIONS),
        HumanMessagePromptTemplate.from_template("{query}"),
    ]
)



from __future__ import annotations

import json
import logging
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, TypeAlias, Union

from langchain_core.documents import Document
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import BaseMessage
from langchain_core.prompts.chat import ChatPromptTemplate
from pymongo import MongoClient, UpdateOne
from pymongo.collection import Collection
from pymongo.errors import OperationFailure
from pymongo.results import BulkWriteResult

from langchain_mongodb.graphrag import example_templates, prompts

from ..utils import DRIVER_METADATA, _append_client_metadata
from .prompts import rag_prompt
from .schema import entity_schema

if TYPE_CHECKING:
    Entity: TypeAlias = Dict[str, Any]
    """Represents an Entity in the knowledge graph with specific schema. See .schema"""

    import holoviews  # type: ignore[import-untyped]
    import networkx

logger = logging.getLogger(__name__)


class MongoDBGraphStore:
    """GraphRAG DataStore

    GraphRAG is a ChatModel that provides responses to semantic queries
    based on a Knowledge Graph that an LLM is used to create.
    As in Vector RAG, we augment the Chat Model's training data
    with relevant information that we collect from documents.

    In Vector RAG, one uses an "Embedding" model that converts both
    the query, and the potentially relevant documents, into vectors,
    which can then be compared, and the most similar supplied to the
    Chat Model as context to the query.

    In Graph RAG, one uses an "Entity-Extraction" model that converts
    text into Entities and their relationships, a Knowledge Graph.
    Comparison is done by Graph traversal, finding entities connected
    to the query prompts. These are then supplied to the Chat Model as context.
    The main difference is that GraphRAG's output is typically in a structured format.

    GraphRAG excels in finding links and common entities,
    even if these come from different articles. It can combine information from
    distinct sources providing richer context than Vector RAG in certain cases.

    Here are a few examples of so-called multi-hop questions where GraphRAG excels:
    - What is the connection between ACME Corporation and GreenTech Ltd.?
    - Who is leading the SolarGrid Initiative, and what is their role?
    - Which organizations are participating in the SolarGrid Initiative?
    - What is John Doe’s role in ACME’s renewable energy projects?
    - Which company is headquartered in San Francisco and involved in the SolarGrid Initiative?

    In Graph RAG, one uses an Entity-Extraction model that interprets
    text documents that it is given and extracting the query,
    and the potentially relevant documents, into graphs. These are
    composed of nodes that are entities (nouns) and edges that are relationships.
    The idea is that the graph can find connections between entities and
    hence answer questions that require more than one connection.

    In MongoDB, Knowledge Graphs are stored in a single Collection.
    Each MongoDB Document represents a single entity (node),
    and its relationships (edges) are defined in a nested field named
    "relationships". The schema, and an example, are described in the
    :data:`~langchain_mongodb.graphrag.prompts.entity_context` prompts module.

    When a query is made, the model extracts the entities in it,
    then traverses the graph to find connections.
    The closest entities and their relationships form the context
    that is included with the query to the Chat Model.

    Consider this example Query: "Does John Doe work at MongoDB?"
    GraphRAG can answer this question even if the following two statements come
    from completely different sources.
    - "Jane Smith works with John Doe."
    - "Jane Smith works at MongoDB."
    """

    def __init__(
        self,
        *,
        connection_string: Optional[str] = None,
        database_name: Optional[str] = None,
        collection_name: Optional[str] = None,
        collection: Optional[Collection] = None,
        entity_extraction_model: BaseChatModel,
        entity_prompt: Optional[ChatPromptTemplate] = None,
        query_prompt: Optional[ChatPromptTemplate] = None,
        max_depth: int = 3,
        allowed_entity_types: Optional[List[str]] = None,
        allowed_relationship_types: Optional[List[str]] = None,
        entity_examples: Optional[str] = None,
        entity_name_examples: str = "",
        validate: bool = False,
        validation_action: str = "warn",
    ):
        """
        Args:
            connection_string: A valid MongoDB connection URI.
            database_name: The name of the database to connect to.
            collection_name: The name of the collection to connect to.
            collection: A Collection that will represent a Knowledge Graph.
                ** One may pass a Collection in lieu of connection_string, database_name, and collection_name.
            entity_extraction_model: LLM for converting documents into Graph of Entities and Relationships.
            entity_prompt: Prompt to fill graph store with entities following schema.
                Defaults to .prompts.ENTITY_EXTRACTION_INSTRUCTIONS
            query_prompt: Prompt extracts entities and relationships as search starting points.
                Defaults to .prompts.NAME_EXTRACTION_INSTRUCTIONS
            max_depth: Maximum recursion depth in graph traversal.
            allowed_entity_types: If provided, constrains search to these types.
            allowed_relationship_types: If provided, constrains search to these types.
            entity_examples: A string containing any number of additional examples to provide as context for entity extraction.
            entity_name_examples: A string appended to prompts.NAME_EXTRACTION_INSTRUCTIONS containing examples.
            validate: If True, entity schema will be validated on every insert or update.
            validation_action: One of {"warn", "error"}.
              - If "warn", the default, documents will be inserted but errors logged.
              - If "error", an exception will be raised if any document does not match the schema.
        """
        self._schema = deepcopy(entity_schema)
        collection_existed = True
        if connection_string and collection is not None:
            raise ValueError(
                "Pass one of: connection_string, database_name, and collection_name"
                "OR a MongoDB Collection."
            )
        if collection is None:  # collection is specified by uri and names
            assert collection_name is not None
            assert database_name is not None
            client: MongoClient = MongoClient(
                connection_string,
                driver=DRIVER_METADATA,
            )
            db = client[database_name]
            if collection_name not in db.list_collection_names(
                authorizedCollections=True
            ):
                validator = {"$jsonSchema": self._schema} if validate else None
                collection = client[database_name].create_collection(
                    collection_name,
                    validator=validator,
                    validationAction=validation_action,
                )
                collection_existed = False
            else:
                collection = db[collection_name]
        else:
            if not isinstance(collection, Collection):
                raise ValueError(
                    "collection must be a MongoDB Collection. "
                    "Consider using connection_string, database_name, and collection_name."
                )

        if validate and collection_existed:
            # first check for existing validator
            collection_info = collection.database.command(
                "listCollections", filter={"name": collection.name}
            )
            collection_options = collection_info.get("cursor", {}).get("firstBatch", [])
            validator = collection_options[0].get("options", {}).get("validator", None)
            if not validator:
                try:
                    collection.database.command(
                        "collMod",
                        collection.name,
                        validator={"$jsonSchema": self._schema},
                        validationAction=validation_action,
                    )
                except OperationFailure:
                    logger.warning(
                        "Validation will NOT be performed. "
                        "User must be DB Admin to add validation **after** a Collection is created. \n"
                        "Please add validator when you create collection: "
                        "db.create_collection.(coll_name, validator={'$jsonSchema': schema.entity_schema})"
                    )
        self.collection = collection

        # append_metadata was added in PyMongo 4.14.0, but is a valid database name on earlier versions
        _append_client_metadata(collection.database.client)

        self.entity_extraction_model = entity_extraction_model
        self.entity_prompt = (
            prompts.entity_prompt if entity_prompt is None else entity_prompt
        )
        self.query_prompt = (
            prompts.query_prompt if query_prompt is None else query_prompt
        )
        self.entity_examples = (
            example_templates.entity_extraction
            if entity_examples is None
            else entity_examples
        )
        self.entity_name_examples = entity_name_examples

        self.max_depth = max_depth
        self._schema = deepcopy(entity_schema)
        if allowed_entity_types:
            self.allowed_entity_types = allowed_entity_types
            self._schema["properties"]["type"]["enum"] = allowed_entity_types  # type:ignore[index]
        else:
            self.allowed_entity_types = []
        if allowed_relationship_types:
            # Update Prompt
            self.allowed_relationship_types = allowed_relationship_types
            # Update schema. Disallow other keys..
            self._schema["properties"]["relationships"]["properties"]["types"][  # type:ignore[index]
                "enum"
            ] = allowed_relationship_types
        else:
            self.allowed_relationship_types = []

    @property
    def entity_schema(self) -> dict[str, Any]:
        """JSON Schema Object of Entities. Will be applied if validate is True.

        See Also:
            `$jsonSchema <https://www.mongodb.com/docs/manual/reference/operator/query/jsonSchema/>`_
        """
        return self._schema

    @classmethod
    def from_connection_string(
        cls,
        connection_string: str,
        database_name: str,
        collection_name: str,
        entity_extraction_model: BaseChatModel,
        entity_prompt: ChatPromptTemplate = prompts.entity_prompt,
        query_prompt: ChatPromptTemplate = prompts.query_prompt,
        max_depth: int = 3,
        allowed_entity_types: Optional[List[str]] = None,
        allowed_relationship_types: Optional[List[str]] = None,
        entity_examples: Optional[str] = None,
        entity_name_examples: str = "",
        validate: bool = False,
        validation_action: str = "warn",
    ) -> MongoDBGraphStore:
        """Construct a `MongoDB KnowLedge Graph for RAG`
        from a MongoDB connection URI.

        Args:
            connection_string: A valid MongoDB connection URI.
            database_name: The name of the database to connect to.
            collection_name: The name of the collection to connect to.
            entity_extraction_model: LLM for converting documents into Graph of Entities and Relationships.
            entity_prompt: Prompt to fill graph store with entities following schema.
            query_prompt: Prompt extracts entities and relationships as search starting points.
            max_depth: Maximum recursion depth in graph traversal.
            allowed_entity_types: If provided, constrains search to these types.
            allowed_relationship_types: If provided, constrains search to these types.
            entity_examples: A string containing any number of additional examples to provide as context for entity extraction.
            entity_name_examples: A string appended to prompts.NAME_EXTRACTION_INSTRUCTIONS containing examples.
            validate: If True, entity schema will be validated on every insert or update.
            validation_action: One of {"warn", "error"}.
              - If "warn", the default, documents will be inserted but errors logged.
              - If "error", an exception will be raised if any document does not match the schema.

        Returns:
            A new MongoDBGraphStore instance.
        """
        client: MongoClient = MongoClient(
            connection_string,
            driver=DRIVER_METADATA,
        )
        collection = client[database_name].create_collection(collection_name)
        return cls(
            collection=collection,
            entity_extraction_model=entity_extraction_model,
            entity_prompt=entity_prompt,
            query_prompt=query_prompt,
            max_depth=max_depth,
            allowed_entity_types=allowed_entity_types,
            allowed_relationship_types=allowed_relationship_types,
            entity_examples=entity_examples,
            entity_name_examples=entity_name_examples,
            validate=validate,
            validation_action=validation_action,
        )

    def close(self) -> None:
        """Close the resources used by the MongoDBGraphStore."""
        self.collection.database.client.close()

    def _write_entities(self, entities: List[Entity]) -> BulkWriteResult | None:
        """Isolate logic to insert and aggregate entities."""
        operations = []
        for entity in entities:
            relationships = entity.get("relationships", {})
            target_ids = relationships.get("target_ids", [])
            types = relationships.get("types", [])
            attributes = relationships.get("attributes", [])

            # Ensure the lengths of target_ids, types, and attributes align
            if not (len(target_ids) == len(types) == len(attributes)):
                logger.warning(
                    f"Targets, types, and attributes do not have the same length for {entity['_id']}!"
                )

            operations.append(
                UpdateOne(
                    filter={"_id": entity["_id"]},  # Match on _id
                    update={
                        "$setOnInsert": {  # Set if upsert
                            "_id": entity["_id"],
                            "type": entity["type"],
                        },
                        "$addToSet": {  # Update without overwriting
                            **{
                                f"attributes.{k}": {"$each": v}
                                for k, v in entity.get("attributes", {}).items()
                            },
                        },
                        "$push": {  # Push new entries into arrays
                            "relationships.target_ids": {"$each": target_ids},
                            "relationships.types": {"$each": types},
                            "relationships.attributes": {"$each": attributes},
                        },
                    },
                    upsert=True,
                )
            )

        # Execute bulk write for the entities
        if operations:
            return self.collection.bulk_write(operations)
        return None

    def add_documents(
        self, documents: Union[Document, List[Document]]
    ) -> List[BulkWriteResult]:
        """Extract entities and upsert into the collection.

        Each entity is represented by a single MongoDB Document.
        Existing entities identified in documents will be updated.

        Args:
            documents: list of textual documents and associated metadata.
        Returns:
            List containing metadata on entities inserted and updated, one value for each input document.
        """
        documents = [documents] if isinstance(documents, Document) else documents
        results = []
        for doc in documents:
            # Call LLM to find all Entities in doc
            entities = self.extract_entities(doc.page_content)
            logger.debug(f"Entities found: {[e['_id'] for e in entities]}")
            # Insert new or combine with existing entities
            new_results = self._write_entities(entities)
            assert new_results is not None
            results.append(new_results)
        return results

    def extract_entities(self, raw_document: str, **kwargs: Any) -> List[Entity]:
        """Extract entities and their relations using chosen prompt and LLM.

        Args:
            raw_document: A single text document as a string. Typically prose.
        Returns:
            List of Entity dictionaries.
        """
        # Combine the LLM with the prompt template to form a chain
        chain = self.entity_prompt | self.entity_extraction_model
        # Invoke on a document to extract entities and relationships
        response = chain.invoke(
            dict(
                input_document=raw_document,
                entity_schema=self.entity_schema,
                entity_examples=self.entity_examples,
                allowed_entity_types=self.allowed_entity_types,
                allowed_relationship_types=self.allowed_relationship_types,
            )
        )
        # Post-Process output string into list of entity json documents
        # Strip the ```json prefix and trailing ```
        assert isinstance(response.content, str)
        json_string = (
            response.content.removeprefix("```json").removesuffix("```").strip()
        )
        extracted = json.loads(json_string)
        return extracted["entities"]

    def extract_entity_names(self, raw_document: str, **kwargs: Any) -> List[str]:
        """Extract entity names from a document for similarity_search.

        The second entity extraction has a different form and purpose than
        the first as we are looking for starting points of our search and
        paths to follow. We aim to find source nodes,  but no target nodes or edges.

        Args:
            raw_document: A single text document as a string. Typically prose.
        Returns:
            List of entity names / _ids.
        """
        # Combine the llm with the prompt template to form a chain
        chain = self.query_prompt | self.entity_extraction_model
        # Invoke on a document to extract entities and relationships
        response = chain.invoke(
            dict(
                input_document=raw_document,
                entity_name_examples=self.entity_name_examples,
                allowed_entity_types=self.allowed_entity_types,
            )
        )
        # Post-Process output string into list of entity json documents
        # Strip the ```json prefix and suffix
        assert isinstance(response.content, str)
        json_string = (
            response.content.removeprefix("```json").removesuffix("```").strip()
        )
        return json.loads(json_string)

    def find_entity_by_name(self, name: str) -> Optional[Entity]:
        """Utility to get Entity dict from Knowledge Graph / Collection.
        Args:
            name: _id string to look for.
        Returns:
            List of Entity dicts if any match name.
        """
        return self.collection.find_one({"_id": name})

    def related_entities(
        self,
        starting_entities: List[str],
        max_depth: Optional[int] = None,
    ) -> List[Entity]:
        """Traverse Graph along relationship edges to find connected entities.

        Args:
            starting_entities: Traversal begins with documents whose _id fields match these strings.
            max_depth: Recursion continues until no more matching documents are found,
                or until the operation reaches a recursion depth specified by this parameter.

        Returns:
            List of connected entities.
        """
        pipeline = [
            # Match starting entities
            {"$match": {"_id": {"$in": starting_entities}}},
            {
                "$graphLookup": {
                    "from": self.collection.name,
                    "startWith": "$relationships.target_ids",  # Start traversal with relationships.target_ids
                    "connectFromField": "relationships.target_ids",  # Traverse via relationships.target_ids
                    "connectToField": "_id",  # Match to entity _id field
                    "as": "connections",  # Store connections
                    "maxDepth": max_depth or self.max_depth,  # Limit traversal depth
                    "depthField": "depth",  # Track depth
                }
            },
            # Exclude connections from the original document
            {
                "$project": {
                    "_id": 0,
                    "original": {
                        "_id": "$_id",
                        "type": "$type",
                        "attributes": "$attributes",
                        "relationships": "$relationships",
                    },
                    "connections": 1,  # Retain connections for deduplication
                }
            },
            # Combine original and connections into one array
            {
                "$project": {
                    "combined": {
                        "$concatArrays": [
                            ["$original"],  # Include original as an array
                            "$connections",  # Include connections
                        ]
                    }
                }
            },
            # Unwind the combined array into individual documents
            {"$unwind": "$combined"},
            # Remove duplicates by grouping on `_id` and keeping the first document
            {
                "$group": {
                    "_id": "$combined._id",  # Group by entity _id
                    "entity": {"$first": "$combined"},  # Keep the first occurrence
                }
            },
            # Format the final output
            {
                "$replaceRoot": {
                    "newRoot": "$entity"  # Use the deduplicated document as the root
                }
            },
        ]
        return list(self.collection.aggregate(pipeline))  # type:ignore[arg-type]

    def similarity_search(self, input_document: str) -> List[Entity]:
        """Retrieve list of connected Entities found via traversal of KnowledgeGraph.

        1. Use LLM & Prompt to find entities within the input_document itself.
        2. Find Entity Nodes that match those found in the input_document.
        3. Traverse the graph using these as starting points.

        Args:
            input_document: String to find relevant documents for.
        Returns:
            List of connected Entity dictionaries.
        """
        starting_ids: List[str] = self.extract_entity_names(input_document)
        return self.related_entities(starting_ids)

    def chat_response(
        self,
        query: str,
        chat_model: Optional[BaseChatModel] = None,
        prompt: Optional[ChatPromptTemplate] = None,
    ) -> BaseMessage:
        """Responds to a query given information found in Knowledge Graph.

        Args:
            query: Prompt before it is augmented by Knowledge Graph.
            chat_model: ChatBot. Defaults to entity_extraction_model.
            prompt: Alternative Prompt Template. Defaults to prompts.rag_prompt.
        Returns:
            Response Message. response.content contains text.
        """
        if chat_model is None:
            chat_model = self.entity_extraction_model
        if prompt is None:
            prompt = rag_prompt

        # Perform Retrieval on knowledge graph
        related_entities = self.similarity_search(query)
        # Combine the LLM with the prompt template to form a chain
        chain = prompt | chat_model
        # Invoke with query
        return chain.invoke(
            dict(
                query=query,
                related_entities=related_entities,
                entity_schema=entity_schema,
            )
        )

    def to_networkx(
        self,
        nx_opts: Optional[dict] = None,
        json_opts: Optional[dict] = None,
        **kwargs: Any,
    ) -> networkx.DiGraph:
        """Utility converts Entity Collection to `NetworkX DiGraph <https://networkx.org/documentation/stable/index.html>`_

        NOTE: Requires optional-dependency "viz", i.e. `pip install "langchain-mongodb[viz]"`.

        Args:
            nx_opts: Keyword arguments for networkx calls.
            json_opts: Keyword arguments for printing of node attributes and types.
            **kwargs: Keyword arguments available for compatibility.

        Returns: networkx.DiGraph
        """

        try:
            import json

            import networkx as nx
        except ImportError as e:
            raise ImportError(
                "Install optional-dependency `viz` for networkx or to view in Holoviews"
            ) from e

        def _safe_get(lst: list, i: int, default: Any = "") -> Any:
            return lst[i] if i < len(lst) else default

        nx_opts = {} if nx_opts is None else nx_opts
        json_opts = {} if json_opts is None else json_opts

        # First pass: Add all nodes with their attributes
        nx_graph = nx.DiGraph(**nx_opts)
        for doc in self.collection.find({}, {"type": 1, "attributes": 1}):
            # Add node with all attributes
            node_id = doc["_id"]
            node_attrs = {}
            node_attrs["type"] = json.dumps(doc.get("type", ""), **json_opts)
            node_attrs["attributes"] = json.dumps(
                doc.get("attributes", {}), **json_opts
            )
            nx_graph.add_node(node_id, **node_attrs, **json_opts)

        # Second pass: Add edges based on relationships
        for doc in self.collection.find({}, {"_id": 1, "relationships": 1}):
            source_id = doc["_id"]
            relationships = doc.get("relationships", {})
            # relationships can contain numerous target_ids, each with type and attributes
            target_ids = relationships.get("target_ids", [])
            n_targets = len(target_ids)
            types = relationships.get("types", [])
            attrs = relationships.get("attributes", [])

            for t in range(n_targets):
                # Add edge and attributes
                edge_attrs = {}
                edge_attrs["type"] = json.dumps(_safe_get(types, t), **json_opts)
                edge_attrs["attributes"] = json.dumps(_safe_get(attrs, t), **json_opts)
                if nx_graph.has_node(target_ids[t]):
                    nx_graph.add_edge(source_id, target_ids[t], **edge_attrs, **nx_opts)
                else:
                    logger.warning(
                        f"{source_id=} references {target_ids[t]=} not found in collection"
                    )

        return nx_graph

    def view(
        self,
        layout: Optional[Callable] = None,
        nx_opts: Optional[dict] = None,
        json_opts: Optional[dict] = None,
        edge_opts: Optional[dict] = None,
        node_opts: Optional[dict] = None,
        **kwargs: Any,
    ) -> holoviews.Graph:
        """Draws a Knowledge Graph as Holoviews/Bokeh interactive plot.

        We first convert the entity collection to a NetworkX Graph,
        and then convert it to a Holoviews Graph via their API.

        The default layout chosen is the spring_layout.
        This maximizes the distance between nodes. As our entities have a type field,
        however, another good layout choice might be
        `layout=nx.multipartite_layout, nx_opts["subset_key"]= "type"`
        as multipartite layout positions nodes in straight lines by subset key.

        NOTE: Requires optional-dependency "viz", i.e. `pip install "langchain-mongodb[viz]"`.

        You can save the view as any HoloViews object with `.save`.
        The type will be inferred from the filename's suffix,
        (e.g., hv.save(graph, "graph.html")) or by clicking the download widget
        on the Bokeh plot from a Jupyter notebook.

        Args:
            layout: `networkx layout. <https://networkx.org/documentation/stable/reference/drawing.html#module-networkx.drawing.layout>`_
                Defaults to networkx.spring_layout.
            nx_opts: Keyword arguments for to_networkx function.
            json_opts: Keyword arguments for printing of node attributes and types.
            edge_opts: Keyword arguments to draw edges.
            node_opts: Keyword arguments to draw nodes.
            **kwargs: Keyword arguments available for compatibility.

        Returns: `holoviews.Graph <https://holoviews.org/user_guide/Network_Graphs.html>`_

        """
        try:
            import holoviews as hv
            import networkx as nx

            hv.extension("bokeh")
        except ImportError as e:
            raise ImportError("To view graph, install optional-dependency `viz`") from e

        hv.opts.defaults(
            hv.opts.Graph(xaxis=None, yaxis=None), hv.opts.Nodes(xaxis=None, yaxis=None)
        )

        if layout is None:
            layout = nx.spring_layout
        # Convert entity collection to NetworkX graph.
        nx_opts = {} if nx_opts is None else nx_opts
        json_opts = {} if json_opts is None else json_opts
        nx_graph = self.to_networkx(**nx_opts, **json_opts)
        # Convert to HoloViews Graph
        hv_graph = hv.Graph.from_networkx(nx_graph, layout, **nx_opts)
        # Display with hover tools over edges and nodes
        edge_opts = {} if edge_opts is None else edge_opts
        node_opts = {} if node_opts is None else node_opts
        return hv_graph.opts(
            inspection_policy="edges", **edge_opts
        ) * hv_graph.nodes.opts(**node_opts)



from time import sleep, time
from typing import Generator, List

import pytest
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from pymongo import MongoClient
from pymongo.collection import Collection

from langchain_mongodb import MongoDBAtlasVectorSearch
from langchain_mongodb.index import (
    create_fulltext_search_index,
    create_vector_search_index,
)
from langchain_mongodb.retrievers import (
    MongoDBAtlasFullTextSearchRetriever,
    MongoDBAtlasHybridSearchRetriever,
)

from ..utils import DB_NAME, PatchedMongoDBAtlasVectorSearch

COLLECTION_NAME = "langchain_test_retrievers"
COLLECTION_NAME_NESTED = "langchain_test_retrievers_nested"
VECTOR_INDEX_NAME = "vector_index"
EMBEDDING_FIELD = "embedding"
PAGE_CONTENT_FIELD = ["text", "keywords"]
PAGE_CONTENT_FIELD_NESTED = "title.text"
SEARCH_INDEX_NAME = "text_index_multi"
SEARCH_INDEX_NAME_NESTED = "text_index_nested"

TIMEOUT = 60.0
INTERVAL = 0.5


@pytest.fixture(scope="module")
def example_documents() -> List[Document]:
    return [
        Document(
            page_content="In 2023, I visited Paris", metadata={"keywords": "MongoDB"}
        ),
        Document(
            page_content="In 2022, I visited New York",
            metadata={"keywords": "Atlas"},
        ),
        Document(
            page_content="In 2021, I visited New Orleans",
            metadata={"keywords": "Search"},
        ),
        Document(
            page_content="Sandwiches are beautiful. Sandwiches are fine.",
            metadata={"keywords": "is awesome"},
        ),
    ]


@pytest.fixture(scope="module")
def collection(client: MongoClient, dimensions: int) -> Collection:
    """A Collection with both a Vector and a Full-text Search Index"""
    if COLLECTION_NAME not in client[DB_NAME].list_collection_names():
        clxn = client[DB_NAME].create_collection(COLLECTION_NAME)
    else:
        clxn = client[DB_NAME][COLLECTION_NAME]

    clxn.delete_many({})

    if not any([VECTOR_INDEX_NAME == ix["name"] for ix in clxn.list_search_indexes()]):
        create_vector_search_index(
            collection=clxn,
            index_name=VECTOR_INDEX_NAME,
            dimensions=dimensions,
            path="embedding",
            similarity="cosine",
            wait_until_complete=TIMEOUT,
        )

    if not any([SEARCH_INDEX_NAME == ix["name"] for ix in clxn.list_search_indexes()]):
        create_fulltext_search_index(
            collection=clxn,
            index_name=SEARCH_INDEX_NAME,
            field=PAGE_CONTENT_FIELD,
            wait_until_complete=TIMEOUT,
        )

    return clxn


@pytest.fixture(scope="module")
def collection_nested(client: MongoClient, dimensions: int) -> Collection:
    """A Collection with both a Vector and a Full-text Search Index"""
    if COLLECTION_NAME_NESTED not in client[DB_NAME].list_collection_names():
        clxn = client[DB_NAME].create_collection(COLLECTION_NAME_NESTED)
    else:
        clxn = client[DB_NAME][COLLECTION_NAME_NESTED]

    clxn.delete_many({})

    if not any([VECTOR_INDEX_NAME == ix["name"] for ix in clxn.list_search_indexes()]):
        create_vector_search_index(
            collection=clxn,
            index_name=VECTOR_INDEX_NAME,
            dimensions=dimensions,
            path="embedding",
            similarity="cosine",
            wait_until_complete=TIMEOUT,
        )

    if not any(
        [SEARCH_INDEX_NAME_NESTED == ix["name"] for ix in clxn.list_search_indexes()]
    ):
        create_fulltext_search_index(
            collection=clxn,
            index_name=SEARCH_INDEX_NAME_NESTED,
            field=PAGE_CONTENT_FIELD_NESTED,
            wait_until_complete=TIMEOUT,
        )

    return clxn


@pytest.fixture(scope="module")
def indexed_vectorstore(
    collection: Collection,
    example_documents: List[Document],
    embedding: Embeddings,
) -> Generator[MongoDBAtlasVectorSearch, None, None]:
    """Return a VectorStore with example document embeddings indexed."""

    vectorstore = PatchedMongoDBAtlasVectorSearch(
        collection=collection,
        embedding=embedding,
        index_name=VECTOR_INDEX_NAME,
        text_key=PAGE_CONTENT_FIELD,
    )

    vectorstore.add_documents(example_documents)

    yield vectorstore

    vectorstore.collection.delete_many({})


@pytest.fixture(scope="module")
def indexed_nested_vectorstore(
    collection_nested: Collection,
    example_documents: List[Document],
    embedding: Embeddings,
) -> Generator[MongoDBAtlasVectorSearch, None, None]:
    """Return a VectorStore with example document embeddings indexed."""

    vectorstore = PatchedMongoDBAtlasVectorSearch(
        collection=collection_nested,
        embedding=embedding,
        index_name=VECTOR_INDEX_NAME,
        text_key=PAGE_CONTENT_FIELD_NESTED,
    )

    vectorstore.add_documents(example_documents)

    yield vectorstore

    vectorstore.collection.delete_many({})


def test_vector_retriever(indexed_vectorstore: PatchedMongoDBAtlasVectorSearch) -> None:
    """Test VectorStoreRetriever"""
    retriever = indexed_vectorstore.as_retriever()

    query1 = "When did I visit France?"
    results = retriever.invoke(query1)
    assert len(results) == 4
    assert "Paris" in results[0].page_content
    assert "MongoDB" == results[0].metadata["keywords"]

    query2 = "When was the last time I visited new orleans?"
    results = retriever.invoke(query2)
    assert "New Orleans" in results[0].page_content
    assert "Search" == results[0].metadata["keywords"]


def test_hybrid_retriever(indexed_vectorstore: PatchedMongoDBAtlasVectorSearch) -> None:
    """Test basic usage of MongoDBAtlasHybridSearchRetriever"""

    retriever = MongoDBAtlasHybridSearchRetriever(
        vectorstore=indexed_vectorstore,
        search_index_name=SEARCH_INDEX_NAME,
        k=3,
    )

    query1 = "When did I visit France?"
    results = retriever.invoke(query1)
    assert len(results) == 3
    assert "Paris" in results[0].page_content

    query2 = "When was the last time I visited new orleans?"
    results = retriever.invoke(query2)
    assert "New Orleans" in results[0].page_content


def test_hybrid_retriever_deprecated_top_k(
    indexed_vectorstore: PatchedMongoDBAtlasVectorSearch,
) -> None:
    """Test basic usage of MongoDBAtlasHybridSearchRetriever"""
    retriever = MongoDBAtlasHybridSearchRetriever(
        vectorstore=indexed_vectorstore,
        search_index_name=SEARCH_INDEX_NAME,
        top_k=3,
    )

    query1 = "When did I visit France?"
    with pytest.warns(DeprecationWarning):
        results = retriever.invoke(query1)
    assert len(results) == 3
    assert "Paris" in results[0].page_content

    query2 = "When was the last time I visited new orleans?"
    with pytest.warns(DeprecationWarning):
        results = retriever.invoke(query2)
    assert "New Orleans" in results[0].page_content


def test_hybrid_retriever_nested(
    indexed_nested_vectorstore: PatchedMongoDBAtlasVectorSearch,
) -> None:
    """Test basic usage of MongoDBAtlasHybridSearchRetriever"""
    retriever = MongoDBAtlasHybridSearchRetriever(
        vectorstore=indexed_nested_vectorstore,
        search_index_name=SEARCH_INDEX_NAME_NESTED,
        k=3,
    )

    query1 = "What did I visit France?"
    results = retriever.invoke(query1)
    assert len(results) == 3
    assert "Paris" in results[0].page_content

    query2 = "When was the last time I visited new orleans?"
    results = retriever.invoke(query2)
    assert "New Orleans" in results[0].page_content


def test_fulltext_retriever(
    indexed_vectorstore: PatchedMongoDBAtlasVectorSearch,
) -> None:
    """Test result of performing fulltext search.

    The Retriever is independent of the VectorStore.
    We use it here only to get the Collection, which we know to be indexed.
    """

    collection: Collection = indexed_vectorstore.collection

    retriever = MongoDBAtlasFullTextSearchRetriever(
        collection=collection,
        search_index_name=SEARCH_INDEX_NAME,
        search_field=PAGE_CONTENT_FIELD,
    )

    # Wait for the search index to complete.
    search_content = dict(
        index=SEARCH_INDEX_NAME,
        wildcard=dict(query="*", path=PAGE_CONTENT_FIELD, allowAnalyzedField=True),
    )
    n_docs = collection.count_documents({})
    t0 = time()
    while True:
        if (time() - t0) > TIMEOUT:
            raise TimeoutError(
                f"Search index {SEARCH_INDEX_NAME} did not complete in {TIMEOUT}"
            )
        cursor = collection.aggregate([{"$search": search_content}])
        if len(list(cursor)) == n_docs:
            break
        sleep(INTERVAL)

    query = "What is MongoDB"
    results = retriever.invoke(query)
    print(results)
    print(list(collection.list_search_indexes()))
    # assert "New Orleans" in results[0].page_content
    assert "MongoDB" in results[0].metadata["keywords"]
    assert "score" in results[0].metadata



from __future__ import annotations

from datetime import datetime
from typing import Type

from bson import ObjectId
from langchain_tests.unit_tests import ToolsUnitTests

from langchain_mongodb.agent_toolkit import MongoDBDatabase
from langchain_mongodb.agent_toolkit.tool import (
    InfoMongoDBDatabaseTool,
    ListMongoDBDatabaseTool,
    QueryMongoDBCheckerTool,
    QueryMongoDBDatabaseTool,
)

from ..utils import FakeLLM, MockClient


class TestQueryMongoDBDatabaseToolUnit(ToolsUnitTests):
    @property
    def tool_constructor(self) -> Type[QueryMongoDBDatabaseTool]:
        return QueryMongoDBDatabaseTool

    @property
    def tool_constructor_params(self) -> dict:
        return dict(db=MongoDBDatabase(MockClient(), "test"))  # type:ignore[arg-type]

    @property
    def tool_invoke_params_example(self) -> dict:
        return dict(query="db.foo.aggregate()")


class TestInfoMongoDBDatabaseToolUnit(ToolsUnitTests):
    @property
    def tool_constructor(self) -> Type[InfoMongoDBDatabaseTool]:
        return InfoMongoDBDatabaseTool

    @property
    def tool_constructor_params(self) -> dict:
        return dict(db=MongoDBDatabase(MockClient(), "test"))  # type:ignore[arg-type]

    @property
    def tool_invoke_params_example(self) -> dict:
        return dict(collection_names="test")


class TestListMongoDBDatabaseToolUnit(ToolsUnitTests):
    @property
    def tool_constructor(self) -> Type[ListMongoDBDatabaseTool]:
        return ListMongoDBDatabaseTool

    @property
    def tool_constructor_params(self) -> dict:
        return dict(db=MongoDBDatabase(MockClient(), "test"))  # type:ignore[arg-type]

    @property
    def tool_invoke_params_example(self) -> dict:
        return dict()


class TestQueryMongoDBCheckerToolUnit(ToolsUnitTests):
    @property
    def tool_constructor(self) -> Type[QueryMongoDBCheckerTool]:
        return QueryMongoDBCheckerTool

    @property
    def tool_constructor_params(self) -> dict:
        return dict(db=MongoDBDatabase(MockClient(), "test"), llm=FakeLLM())  # type:ignore[arg-type]

    @property
    def tool_invoke_params_example(self) -> dict:
        return dict(query="db.foo.aggregate()")


def test_database_parse_command() -> None:
    db = MongoDBDatabase(MockClient(), "test")  # type:ignore[arg-type]

    command = """db.user.aggregate([ { "$match": { "_id": ObjectId("123412341234123412341234") } } ])"""
    result = db._parse_command(command)
    assert isinstance(result[0]["$match"]["_id"], ObjectId)

    command = """db.user.aggregate([ { "$match": { "date":  ISODate("2017-04-27T04:26:42.709Z") } } ])"""
    result = db._parse_command(command)
    assert isinstance(result[0]["$match"]["date"], datetime)

    command = """db.user.aggregate([ { "$match": { "date": new Date("2017-04-27T04:26:42.709Z") } } ])"""
    result = db._parse_command(command)
    assert isinstance(result[0]["$match"]["date"], datetime)



# Based on https://github.com/langchain-ai/langchain/blob/edbe7d5f5e0dcc771c1f53a49bb784a3960ce448/libs/community/langchain_community/document_loaders/mongodb.py
from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional, Sequence

from langchain_community.document_loaders.base import BaseLoader
from langchain_core.documents import Document
from langchain_core.runnables.config import run_in_executor
from pymongo import MongoClient
from pymongo.collection import Collection

from langchain_mongodb.utils import DRIVER_METADATA, _append_client_metadata

logger = logging.getLogger(__name__)


class MongoDBLoader(BaseLoader):
    """Document Loaders are classes to load Documents.

    Document Loaders are usually used to load a lot of Documents in a single run."""

    def __init__(
        self,
        collection: Collection,
        *,
        filter_criteria: Optional[Dict] = None,
        field_names: Optional[Sequence[str]] = None,
        metadata_names: Optional[Sequence[str]] = None,
        include_db_collection_in_metadata: bool = True,
    ) -> None:
        """
        Initializes the MongoDB loader with necessary database connection
        details and configurations.

        Args:
            collection (Collection): The pymongo collection to fetch documents from.
            filter_criteria (Optional[Dict]): MongoDB filter criteria for querying
            documents.
            field_names (Optional[Sequence[str]]): List of field names to retrieve
            from documents.
            metadata_names (Optional[Sequence[str]]): Additional metadata fields to
            extract from documents.
            include_db_collection_in_metadata (bool): Flag to include database and
            collection names in metadata.
        """
        self.collection = collection
        self.db = collection.database
        self.db_name = self.db.name
        self.collection_name = collection.name
        self.field_names = field_names or []
        self.filter_criteria = filter_criteria or {}
        self.metadata_names = metadata_names or []
        self.include_db_collection_in_metadata = include_db_collection_in_metadata

        # append_metadata was added in PyMongo 4.14.0, but is a valid database name on earlier versions
        _append_client_metadata(self.db.client)

    @classmethod
    def from_connection_string(
        cls,
        connection_string: str,
        db_name: str,
        collection_name: str,
        *,
        filter_criteria: Optional[Dict] = None,
        field_names: Optional[Sequence[str]] = None,
        metadata_names: Optional[Sequence[str]] = None,
        include_db_collection_in_metadata: bool = True,
    ) -> MongoDBLoader:
        """
        Creates a MongoDB loader with necessary database connection
        details and configurations.

        Args:
            connection_string (str): MongoDB connection URI.
            db_name (str):Name of the database to connect to.
            collection_name (str): Name of the collection to fetch documents from.
            filter_criteria (Optional[Dict]): MongoDB filter criteria for querying
            documents.
            field_names (Optional[Sequence[str]]): List of field names to retrieve
            from documents.
            metadata_names (Optional[Sequence[str]]): Additional metadata fields to
            extract from documents.
            include_db_collection_in_metadata (bool): Flag to include database and
            collection names in metadata.
        """
        client: MongoClient[dict[str, Any]] = MongoClient(
            connection_string,
            driver=DRIVER_METADATA,
        )
        collection = client[db_name][collection_name]
        return MongoDBLoader(
            collection,
            filter_criteria=filter_criteria,
            field_names=field_names,
            metadata_names=metadata_names,
            include_db_collection_in_metadata=include_db_collection_in_metadata,
        )

    def close(self) -> None:
        """Close the resources used by the MongoDBLoader."""
        self.db.client.close()

    def load(self) -> List[Document]:
        """Load data into Document objects."""
        result = []
        total_docs = self.collection.count_documents(self.filter_criteria)

        projection = self._construct_projection()

        for doc in self.collection.find(self.filter_criteria, projection):
            metadata = self._extract_fields(doc, self.metadata_names, default="")

            # Optionally add database and collection names to metadata
            if self.include_db_collection_in_metadata:
                metadata.update(
                    {"database": self.db_name, "collection": self.collection_name}
                )

            # Extract text content from filtered fields or use the entire document
            if self.field_names is not None:
                fields = self._extract_fields(doc, self.field_names, default="")
                texts = [str(value) for value in fields.values()]
                text = " ".join(texts)
            else:
                text = str(doc)

            result.append(Document(page_content=text, metadata=metadata))

        if len(result) != total_docs:
            logger.warning(
                f"Only partial collection of documents returned. "
                f"Loaded {len(result)} docs, expected {total_docs}."
            )

        return result

    async def aload(self) -> List[Document]:
        """Asynchronously loads data into Document objects."""
        return await run_in_executor(None, self.load)

    def _construct_projection(self) -> Optional[Dict]:
        """Constructs the projection dictionary for MongoDB query based
        on the specified field names and metadata names."""
        field_names = list(self.field_names) or []
        metadata_names = list(self.metadata_names) or []
        all_fields = field_names + metadata_names
        return {field: 1 for field in all_fields} if all_fields else None

    def _extract_fields(
        self,
        document: Dict,
        fields: Sequence[str],
        default: str = "",
    ) -> Dict:
        """Extracts and returns values for specified fields from a document."""
        extracted = {}
        for field in fields or []:
            value = document
            for key in field.split("."):
                value = value.get(key, default)
                if value == default:
                    break
            new_field_name = field.replace(".", "_")
            extracted[new_field_name] = value
        return extracted



from .database import MongoDBDatabase
from .prompt import MONGODB_AGENT_SYSTEM_PROMPT
from .toolkit import MongoDBDatabaseToolkit

__all__ = ["MongoDBDatabaseToolkit", "MongoDBDatabase", "MONGODB_AGENT_SYSTEM_PROMPT"]



from typing import List

from langchain_core.callbacks.manager import CallbackManagerForRetrieverRun
from langchain_core.documents import Document
from langchain_core.retrievers import BaseRetriever

from langchain_mongodb.graphrag.graph import MongoDBGraphStore


class MongoDBGraphRAGRetriever(BaseRetriever):
    """RunnableSerializable API of MongoDB GraphRAG."""

    graph_store: MongoDBGraphStore
    """Underlying Knowledge Graph storing entities and their relationships."""

    def _get_relevant_documents(
        self, query: str, *, run_manager: CallbackManagerForRetrieverRun
    ) -> List[Document]:
        """Retrieve list of Entities found via traversal of KnowledgeGraph.

        Each Document's page_content is a string representation of the Entity dict.

        Description and details are provided in the underlying Entity Graph:
        :class:`~langchain_mongodb.graphrag.graph.MongoDBGraphStore`

        Args:
            query: String to find relevant documents for
            run_manager: The callback handler to use if desired
        Returns:
            List of relevant documents.
        """
        return [Document(str(e)) for e in self.graph_store.similarity_search(query)]
